#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **class factory** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.cls.utilclsmake` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_make_type() -> None:
    '''
    Test the :func:`beartype._util.cls.utilclsmake.make_type` tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilTypeException
    from beartype._util.cls.utilclsmake import make_type
    from beartype._util.cls.utilclstest import is_type_subclass
    from pytest import raises

    # ....................{ LOCALS                         }....................
    class TooEnamouredOfThatVoice(object):
        '''
        Arbitrary base class from which to subclass classes defined below.
        '''

    # ....................{ PASS                           }....................
    # Assert this factory creates and returns the expected class when the passed
    # name is a valid unqualified Python identifier.
    AndSilence = make_type(type_name='AndSilence')
    assert isinstance(AndSilence, type)
    assert AndSilence.__name__ == 'AndSilence'

    # Assert this factory creates and returns the expected class when passed a
    # tuple of one or more base classes.
    BySolemnVision = make_type(
        type_name='BySolemnVision', type_bases=(TooEnamouredOfThatVoice,))
    assert is_type_subclass(BySolemnVision, TooEnamouredOfThatVoice)

    # Assert this factory creates and returns the expected class when passed a
    # dictionary of one or more class attributes.
    BrightSilverDream = make_type(type_name='BrightSilverDream', type_scope={
        'every_sight': 'His infancy was nurtured.',
        'and_sound': lambda self: self.every_sight,
    })
    assert (
        BrightSilverDream.every_sight ==
        BrightSilverDream().and_sound() == 'His infancy was nurtured.'
    )

    # Assert this factory creates and returns the expected class when the passed
    # module name is a valid Python identifier.
    SentToHisHeart = make_type(
        type_name='SentToHisHeart', type_module_name='its_choicest.impulses')
    assert SentToHisHeart.__module__ == 'its_choicest.impulses'

    # Assert this factory creates and returns the expected class when passed a
    # docstring.
    TheFountains = make_type(
        type_name='TheFountains', type_doc='of divine philosophy')
    assert TheFountains.__doc__ == 'of divine philosophy'

    # ....................{ FAIL                           }....................
    # Assert this factory raises the expected exception when the passed name is
    # the empty string.
    with raises(_BeartypeUtilTypeException):
        make_type('')

    # Assert this factory raises the expected exception when the passed name is
    # *NOT* a valid Python identifier.
    with raises(_BeartypeUtilTypeException):
        make_type('And wasted for fond love of his wild eyes.')

    # Assert this factory raises the expected exception when the passed name is
    # a valid Python identifier that is fully-qualified (i.e., contains one or
    # more "." characters).
    with raises(_BeartypeUtilTypeException):
        make_type('The_fire.of.those_soft_orbs.has_ceased.to_burn')

    # Assert this factory raises the expected exception when the passed module
    # name is the empty string.
    with raises(_BeartypeUtilTypeException):
        make_type(type_name='And_Silence', type_module_name='')

    # Assert this factory raises the expected exception when the passed module
    # name is *NOT* a valid Python identifier.
    with raises(_BeartypeUtilTypeException):
        make_type(
            type_name='And_Silence',
            type_module_name='Locks its mute music in her rugged cell.',
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`328`-compliant **relative module name utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.module.pep.modpep328` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_canonicalize_pep328_module_name_relative() -> None:
    '''
    Test the :func:`beartype._util.module.pep.modpep328.canonicalize_pep328_module_name_relative`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilModulePep328Exception
    from beartype._util.module.pep.modpep328 import (
        canonicalize_pep328_module_name_relative)
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # List of the unqualified basenames comprising the fully-qualified name of
    # the originating module to which the relative module tested below is
    # relative.
    MODULE_BASENAMES_ABSOLUTE = [
        'package_name', 'subpackage_name', 'submodule_name']

    # ....................{ PASS                           }....................
    # Assert this canonicalizer returns the expected strings when passed various
    # PEP 328-compliant relative heights *WITHOUT* suffixing module names
    # relative to various package levels.
    assert canonicalize_pep328_module_name_relative(
        module_name_relative='.',
        module_basenames_absolute=MODULE_BASENAMES_ABSOLUTE,
    ) == 'package_name.subpackage_name'
    assert canonicalize_pep328_module_name_relative(
        module_name_relative='..',
        module_basenames_absolute=MODULE_BASENAMES_ABSOLUTE,
    ) == 'package_name'

    # Assert this canonicalizer returns the expected string when passed a PEP
    # 328-compliant relative module name relative to a single package level.
    assert canonicalize_pep328_module_name_relative(
        module_name_relative='.submodule_sibling_name',
        module_basenames_absolute=MODULE_BASENAMES_ABSOLUTE,
    ) == 'package_name.subpackage_name.submodule_sibling_name'

    # Assert this canonicalizer returns the expected string when passed a PEP
    # 328-compliant relative module name relative to two or more package levels.
    assert canonicalize_pep328_module_name_relative(
        module_name_relative='..submodule_nephew_name',
        module_basenames_absolute=MODULE_BASENAMES_ABSOLUTE,
    ) == 'package_name.submodule_nephew_name'

    # ....................{ FAIL                           }....................
    # Assert this canonicalizer passed a PEP 328-noncompliant absolute module
    # name raises the expected exception.
    with raises(_BeartypeUtilModulePep328Exception):
        canonicalize_pep328_module_name_relative(
            module_name_relative='package_name.subpackage_name.submodule_name',
            module_basenames_absolute=MODULE_BASENAMES_ABSOLUTE,
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **path factories** (i.e., low-level callables creating and returning
:class:`pathlib.Path` instances encapsulating testing-specific paths).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.pytroar import BeartypeTestPathException
from pathlib import Path

# ....................{ RELATIVIZERS                       }....................
# These functions are explicitly camelcased to enable their future refactoring
# into actual classes.

def DirRelative(parent_dir: Path, relative_dirname: str) -> Path:
    '''
    Concrete platform-agnostic :class:`.Path` object encapsulating the absolute
    dirname of a directory relative to the passed :class:`.Path` object
    encapsulating an arbitrary directory if found *or* raise an exception
    otherwise.

    Parameters
    ----------
    parent_dir : Path
        :class:`.Path` encapsulating an arbitrary **parent directory** (i.e.,
        directory containing the subdirectory to be returned).
    relative_dirname : str
        Relative dirname of this subdirectory relative to this parent
        directory.

    Returns
    -------
    Path
        :class:`.Path` directory relative to the passed :class:`.Path` directory.

    Raises
    ------
    BeartypeTestPathException
        If this path exists but is either:

        * *Not* a directory.
        * A directory *not* satisfying the expected filesystem structure.
    OSError
        If either:

        * This path does *not* exist.
        * This path exists but whose resolution to a physical path requires
          resolving one or more cyclic symbolic links inducing an infinite loop.
    '''
    assert isinstance(parent_dir, Path), (
        f'{repr(parent_dir)} not "pathlib.Path" object.')
    assert isinstance(relative_dirname, str), (
        f'{repr(relative_dirname)} not string.')

    # Path encapsulating the relative dirname of this subdirectory relative to
    # this parent directory, which has yet to be validated.
    subdir_unresolved = parent_dir / relative_dirname

    # Canonicalize this relative dirname into an absolute dirname if this path
    # exists *OR* raise an "OSError" exception.
    subdir = subdir_unresolved.resolve(strict=True)

    # If this path is *NOT* a directory, raise an exception.
    if not subdir.is_dir():
        raise BeartypeTestPathException(f'Directory "{subdir}" not found.')
    # Else, this path is a directory.

    # Return this directory.
    return subdir


def FileRelative(parent_dir: Path, relative_filename: str) -> Path:
    '''
    Concrete platform-agnostic :class:`.Path` object encapsulating the absolute
    filename of a file relative to the passed :class:`.Path` object encapsulating
    an arbitrary directory if found *or* raise an exception otherwise.

    Parameters
    ----------
    parent_dir : Path
        :class:`.Path` encapsulating an arbitrary **parent directory** (i.e.,
        directory containing the file to be returned).
    relative_filename : str
        Relative filename of this file relative to this parent directory.

    Returns
    -------
    Path
        :class:`.Path` file relative to the passed :class:`.Path` directory.

    Raises
    ------
    BeartypeTestPathException
        If this path exists but is either:

        * *Not* a directory.
        * A directory *not* satisfying the expected filesystem structure.
    OSError
        If either:

        * This path does *not* exist.
        * This path exists but whose resolution to a physical path requires
          resolving one or more cyclic symbolic links inducing an infinite loop.
    '''
    assert isinstance(parent_dir, Path), (
        f'{repr(parent_dir)} not "pathlib.Path" object.')
    assert isinstance(relative_filename, str), (
        f'{repr(relative_filename)} not string.')

    # Path encapsulating the relative filename of this file relative to this
    # parent directory, which has yet to be validated.
    file_unresolved = parent_dir / relative_filename

    # Canonicalize this relative filename into an absolute filename if this
    # path exists *OR* raise an "OSError" exception.
    file = file_unresolved.resolve(strict=True)

    # If this path is *NOT* a file, raise an exception.
    if not file.is_file():
        raise BeartypeTestPathException(f'File "{file}" not found.')
    # Else, this path is a file.

    # Return this file.
    return file



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **mapping singletons** (i.e., dictionaries commonly required
throughout this codebase, reducing space and time consumption by preallocating
widely used dictionary-centric objects).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._util.kind.maplike.utilmapfrozen import FrozenDict

# ....................{ DICTS                              }....................
FROZENDICT_EMPTY = FrozenDict()
'''
**Empty frozen dictionary** (i.e., :class:`.FrozenDict` object containing *no*
key-value pairs).

Whereas Python guarantees the **empty tuple** (i.e., ``()``) to be a singleton,
Python does *not* extend that guarantee to dictionaries. This empty dictionary
singleton amends that oversight, providing efficient reuse of empty
dictionaries: e.g.,

.. code-block:: pycon

   >>> () is ()
   True  # <-- good. this is good.
   >>> {} is {}
   False  # <-- bad. this is bad.
   >>> from beartype._data.kind.datakindmap import FROZENDICT_EMPTY
   >>> FROZENDICT_EMPTY is FROZENDICT_EMPTY
   True  # <-- good. this is good, because we made it so.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **string joining utilities** (i.e., callables joining passed
strings into new strings delimited by passed substring delimiters).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.typing.datatyping import (
    BoolTristate,
    IterableStrs,
    IterableTypes,
)
from collections.abc import (
    Iterable as IterableABC,
    Sequence as SequenceABC,
)

# ....................{ JOINERS                            }....................
#FIXME: Unit test the "is_double_quoted" parameter, please.
def join_delimited(
    # Mandatory parameters.
    strs: IterableStrs,

    # Mandatory keyword-only parameters.
    *,
    delimiter_if_two: str,
    delimiter_if_three_or_more_nonlast: str,
    delimiter_if_three_or_more_last: str,

    # Optional keyword-only parameters.
    is_double_quoted: bool = False,
) -> str:
    '''
    Concatenate the passed iterable of zero or more strings delimited by the
    passed delimiter (conditionally depending on both the length of this
    sequence and index of each string in this sequence), yielding a
    human-readable string listing arbitrarily many substrings.

    Specifically, this function returns either:

    * If this iterable contains no strings, the empty string.
    * If this iterable contains one string, this string as is is unmodified.
    * If this iterable contains two strings, these strings delimited by the
      passed ``delimiter_if_two`` delimiter.
    * If this iterable contains three or more strings, a string listing these
      contained strings such that:

      * All contained strings except the last two are suffixed by the passed
        ``delimiter_if_three_or_more_nonlast`` delimiter.
      * The last two contained strings are delimited by the passed
        ``delimiter_if_three_or_more_last`` separator.

    Parameters
    ----------
    strs : Iterable[str]
        Iterable of all strings to be joined.
    delimiter_if_two : str
        Substring separating each string contained in this iterable if this
        iterable contains exactly two strings.
    delimiter_if_three_or_more_nonlast : str
        Substring separating each string *except* the last two contained in
        this iterable if this iterable contains three or more strings.
    delimiter_if_three_or_more_last : str
        Substring separating each string the last two contained in this
        iterable if this iterable contains three or more strings.
    is_double_quoted : bool, optional
        :data:`True` only if **double-quoting** (i.e., both prefixing and
        suffixing by the ``"`` character) each item of this iterable. Defaults
        to :data:`False`.

    Returns
    -------
    str
        Concatenation of these strings.

    Examples
    --------
        >>> join_delimited(
        ...     strs=('Fulgrim', 'Perturabo', 'Angron', 'Mortarion'),
        ...     delimiter_if_two=' and ',
        ...     delimiter_if_three_or_more_nonlast=', ',
        ...     delimiter_if_three_or_more_last=', and ',
        ... )
        'Fulgrim, Perturabo, Angron, and Mortarion'
    '''
    assert isinstance(strs, IterableABC) and not isinstance(strs, str), (
        f'{repr(strs)} not non-string iterable.')
    assert isinstance(delimiter_if_two, str), (
        f'{repr(delimiter_if_two)} not string.')
    assert isinstance(delimiter_if_three_or_more_nonlast, str), (
        f'{repr(delimiter_if_three_or_more_nonlast)} not string.')
    assert isinstance(delimiter_if_three_or_more_last, str), (
        f'{repr(delimiter_if_three_or_more_last)} not string.')

    # If this iterable is *NOT* a sequence, internally coerce this iterable
    # into a sequence for subsequent indexing purposes.
    if not isinstance(strs, SequenceABC):
        strs = tuple(strs)
    # Else, this iterable is already a sequence.
    #
    # In either case, this iterable is now a sequence.

    # If double-quoting these strings, do so.
    if is_double_quoted:
        strs = tuple(f'"{text}"' for text in strs)
    # Else, preserve these strings as is.

    # Number of strings in this sequence.
    num_strs = len(strs)

    # If no strings are passed, return the empty string.
    if num_strs == 0:
        return ''
    # If one string is passed, return this string as is.
    elif num_strs == 1:
        # This is clearly a string, yet mypy thinks it's Any
        return strs[0]  # type: ignore[no-any-return]
    # If two strings are passed, return these strings delimited appropriately.
    elif num_strs == 2:
        return f'{strs[0]}{delimiter_if_two}{strs[1]}'
    # Else, three or more strings are passed.

    # All such strings except the last two, delimited appropriately.
    strs_nonlast = delimiter_if_three_or_more_nonlast.join(strs[0:-2])

    # The last two such strings, delimited appropriately.
    strs_last = f'{strs[-2]}{delimiter_if_three_or_more_last}{strs[-1]}'

    # Return these two substrings, delimited appropriately.
    return f'{strs_nonlast}{delimiter_if_three_or_more_nonlast}{strs_last}'

# ....................{ JOINERS ~ conjunction              }....................
#FIXME: Unit test us up, please.
def join_delimited_conjunction(strs: IterableStrs, **kwargs) -> str:
    '''
    Concatenate the passed iterable of zero or more strings delimited by commas
    and/or the conjunction "and" (conditionally depending on both the length of
    this iterable and index of each string in this iterable), yielding a
    human-readable string listing arbitrarily many substrings conjunctively.

    Specifically, this function returns either:

    * If this iterable contains no strings, the empty string.
    * If this iterable contains one string, this string as is is unmodified.
    * If this iterable contains two strings, these strings delimited by the
      conjunction "and".
    * If this iterable contains three or more strings, a string listing these
      contained strings such that:

      * All contained strings except the last two are suffixed by commas.
      * The last two contained strings are delimited by the conjunction "and".

    Parameters
    ----------
    strs : Iterable[str]
        Iterable of all strings to be concatenated conjunctively.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`.join_delimeted` function underlying this higher-level function.

    Returns
    -------
    str
        Conjunctive concatenation of these strings.
    '''

    # One of us. We accept one-liner. One of us.
    return join_delimited(
        strs=strs,
        delimiter_if_two=' and ',
        delimiter_if_three_or_more_nonlast=', ',
        delimiter_if_three_or_more_last=', and ',
        **kwargs
    )

# ....................{ JOINERS ~ disjunction              }....................
def join_delimited_disjunction(strs: IterableStrs, **kwargs) -> str:
    '''
    Concatenate the passed iterable of zero or more strings delimited by commas
    and/or the disjunction "or" (conditionally depending on both the length of
    this iterable and index of each string in this iterable), yielding a
    human-readable string listing arbitrarily many substrings disjunctively.

    Specifically, this function returns either:

    * If this iterable contains no strings, the empty string.
    * If this iterable contains one string, this string as is is unmodified.
    * If this iterable contains two strings, these strings delimited by the
      disjunction "or".
    * If this iterable contains three or more strings, a string listing these
      contained strings such that:

      * All contained strings except the last two are suffixed by commas.
      * The last two contained strings are delimited by the disjunction "or".

    Parameters
    ----------
    strs : Iterable[str]
        Iterable of all strings to be concatenated disjunctively.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`.join_delimeted` function underlying this higher-level function.

    Returns
    -------
    str
        Disjunctive concatenation of these strings.
    '''

    # He will join us... OR DIE! *cackling heard*
    return join_delimited(
        strs=strs,
        delimiter_if_two=' or ',
        delimiter_if_three_or_more_nonlast=', ',
        delimiter_if_three_or_more_last=', or ',
        **kwargs
    )


def join_delimited_disjunction_types(
    # Mandatory parameters.
    types: IterableTypes,

    # Optional parameters.
    is_color: BoolTristate = False,
) -> str:
    '''
    Concatenate the human-readable classname of each class in the passed
    iterable delimited by commas and/or the disjunction "or" (conditionally
    depending on both the length of this iterable and index of each string in
    this iterable), yielding a human-readable string listing arbitrarily many
    classnames disjunctively.

    Parameters
    ----------
    types : Iterable[type]
        Iterable of all classes whose human-readable classnames are to be
        concatenated disjunctively.
    is_color : BoolTristate, optional
        Tri-state colouring boolean governing ANSI usage. See the
        :attr:`beartype.BeartypeConf.is_color` attribute for further details.
        Defaults to :data:`False`.

    Returns
    -------
    str
        Disjunctive concatenation of these classnames.
    '''

    # Avoid circular import dependencies.
    from beartype._util.text.utiltextlabel import label_type

    # Make it so, ensign.
    return join_delimited_disjunction(
        label_type(cls=cls, is_color=is_color) for cls in types)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **current package beartype import hooked submodule** (i.e., data
module mimicking real-world usage of the
:func:`beartype.claw.beartype_this_package` import hook from an arbitrarily
deeply nested submodule ``{some_package}...{some_submodule}` of an arbitrary
third-party package ``{some_package}``).
'''

# ....................{ IMPORTS                            }....................
from beartype import beartype
from beartype.roar import (
    BeartypeCallHintParamViolation,
    BeartypeCallHintReturnViolation,
    BeartypeDoorHintViolation,
)
from beartype.typing import (
    # Iterable,
    List,
    Optional,
    Union,
)
from pytest import raises

#FIXME: Fascinatingly, this doesn't actually test anything. Why? We have *NO*
#idea whatsoever, but strongly suspect that pytest's own import hooks are
#silently conflicting with @beartype's in a manner preventing @beartype from
#guarding against regressions in this functionality. For now, the *ONLY* means
#of testing this is to do so manually in an external module subject to a
#beartype import hook that performs this import. Gag me with a spork.
# Intentionally import a pure-Python module from the standard library that
# internally imports a C extension -- in this case, "_struct". Doing so ensures
# that beartype import hooks properly support loading of C extensions, which our
# initial implementation did *NOT*. Avoid regressions by testing this, please.
import struct

# from beartype.claw._importlib.clawimpcache import module_name_to_beartype_conf
# print(f'this_submodule conf: {repr(module_name_to_beartype_conf)}')

# ....................{ PEP 526                            }....................
# Validate that the beartype_this_package() import hook installed by the parent
# "beartype_this_package.__init__" submodule implicitly appends all PEP
# 526-compliant annotated assignment statements in other submodules of this
# "beartype_this_package" subpackage with calls to beartype's statement-level
# beartype.door.die_if_unbearable() exception-raiser.

# Assert that a PEP 526-compliant assignment statement assigning an object
# satisfying the type hint annotating that statement raises *NO* exception.
#
# Note that this type hint is intentionally annotated as "float" rather than
# either "int" *OR* "Union[int, float]", exercising that that import hook
# successfully associated this and *ALL* other submodules of the
# "beartype_this_package" subpackage with a non-default beartype configuration
# enabling the PEP 484-compliant implicit numeric tower (i.e.,
# "is_pep484_tower=True").
loves_philosophy: float = len('The fountains mingle with the river')

# Assert that a PEP 526-compliant assignment statement assigning an object
# violating the type hint annotating that statement raises the expected
# exception.
with raises(BeartypeDoorHintViolation):
    and_the_rivers_with_the_ocean: List[str] = (
        'The winds of heaven mix for ever')

# ....................{ FUNCTIONS ~ accept                 }....................
# Functions whose type-checking implicitly accepts the non-default beartype
# configuration established by the sibling "__init__" submodule in this
# subpackage.

def nothing_in_the_world(is_single: Union[float, bytes]) -> Optional[complex]:
    '''
    Arbitrary function either returning the passed integer first doubled and
    then coerced into a complex number with imaginary component ``1`` if this
    integer is non-zero *or* raising a :exc:`.BeartypeCallHintParamViolation`
    exception otherwise (i.e., if this integer is zero), exercising that
    beartype import hooks decorate global functions as expected.
    '''

    def in_one_spirit(meet_and_mingle: Optional[float]) -> Union[complex, str]:
        '''
        Arbitrary closure either returning the passed integer first doubled and
        then coerced into a complex number with imaginary component ``1`` if
        this integer is non-:data:`None` *or* raising a
        :exc:`.BeartypeCallHintReturnViolation` exception otherwise (i.e., if
        this integer is :data:`None`), exercising that beartype import hooks
        decorate closures as expected.
        '''

        # Return either...
        return (
            # If this integer is "None", explicitly force this
            # @beartype-decorated closure to raise a
            # "BeartypeCallHintReturnViolation" exception;
            None
            if meet_and_mingle is None else
            # Else, this integer is non-"None". In this case, return this
            # integer doubled and then coerced into a complex number with
            # imaginary component "1". Why? Just because. *THIS IS BEARTYPE.*
            is_single + meet_and_mingle + 1j
        )

    # Return either...
    return (
        # If this integer is zero, explicitly force this @beartype-decorated
        # closure to raise a "BeartypeCallHintReturnViolation" exception.
        in_one_spirit(None)
        if is_single == 0 else
        # Else, this integer is non-zero. In this case, return this integer
        # doubled and then coerced into a complex number with imaginary
        # component "1". Why? Just because. *THIS IS BEARTYPE.* Graaaah!
        in_one_spirit(is_single)
    )

# Assert that calling this function passed an arbitrary integer returns the
# expected complex number *WITHOUT* raising an exception.
assert nothing_in_the_world(len('Why not I with thine?')) == 42 + 1j

# Assert that calling this function passed an invalid parameter raises the
# expected exception.
with raises(BeartypeCallHintParamViolation):
    nothing_in_the_world('See the mountains kiss high heaven')

# Assert that calling this function passed zero raises the expected exception
# from the closure defined and called by this function.
with raises(BeartypeCallHintReturnViolation):
    nothing_in_the_world(len('And the waves') - len('clasp another'))

# ....................{ FUNCTIONS ~ reject                 }....................
# Functions whose type-checking explicitly rejects the non-default beartype
# configuration established by the sibling "__init__" submodule in this
# subpackage (i.e., due to being decorated by the @beartype decorator implicitly
# configured by the default beartype configuration). In particular, these
# functions disable the PEP 484-compliant implicit numeric tower such that:
# * "float" just means "float" *WITHOUT* being expanded to "float | int".
# * "complex" just means "complex" *WITHOUT* being expanded to "complex | float
#   | int".

@beartype  # <-- prefer the default beartype configuration
def alastor_or(the_spirit_of_solitude: Union[float, bytes]) -> (
    Optional[complex]):
    '''
    Arbitrary function either returning the passed float first doubled and then
    coerced into a complex number with imaginary component ``1`` if this float
    is non-zero *or* raising a :exc:`.BeartypeCallHintParamViolation` exception
    otherwise (i.e., if this float is zero), exercising that beartype import
    hooks decorate global functions as expected.
    '''

    @beartype  # <-- prefer the default beartype configuration
    def nondum_amabam(et_amare_amabam: Optional[float]) -> Union[complex, str]:
        '''
        Arbitrary closure either returning the passed float first doubled and
        then coerced into a complex number with imaginary component ``1`` if
        this float is non-:data:`None` *or* raising a
        :exc:`.BeartypeCallHintReturnViolation` exception otherwise (i.e., if
        this float is :data:`None`), exercising that beartype import hooks
        decorate closures as expected.
        '''

        # Return either...
        return (
            # If this float is "None", explicitly force this
            # @beartype-decorated closure to raise a
            # "BeartypeCallHintReturnViolation" exception;
            None
            if et_amare_amabam is None else
            # Else, this float is non-"None". In this case, return this float
            # doubled and then coerced into a complex number with imaginary
            # component "1". Why? Just because. *THIS IS BEARTYPE.* Graaaah!
            the_spirit_of_solitude + et_amare_amabam + 1j
        )

    # Return either...
    return (
        # If this float is zero, explicitly force this @beartype-decorated
        # closure to raise a "BeartypeCallHintReturnViolation" exception.
        nondum_amabam(None)
        if the_spirit_of_solitude == 0.0 else
        # Else, this integer is non-zero. In this case, return this integer
        # doubled and then coerced into a complex number with imaginary
        # component "1". Why? Just because. *THIS IS BEARTYPE.* Graaaah!
        nondum_amabam(the_spirit_of_solitude)
    )


# Assert that calling this function passed an arbitrary float returns the
# expected complex number *WITHOUT* raising an exception.
assert alastor_or(len('Earth, ocean, air, beloved brotherhood!') + 0.0) == (
    78 + 1j)

# Assert that calling this function passed an invalid parameter raises the
# expected exception.
with raises(BeartypeCallHintParamViolation):
    alastor_or(len('If our great Mother has imbued my soul'))

# Assert that calling this function passed zero raises the expected exception
# from the closure defined and called by this function.
with raises(BeartypeCallHintReturnViolation):
    alastor_or(len('quid amarem') - len('amans amare') + 0.0)

# ....................{ FUNCTIONS ~ signature              }....................
# Functions whose signatures exhibit various uncommon edge cases preventing
# regressions in the "beartype.claw" API.

def of_desperate_hope(*, wrinkled_his, quivering_lips: str):
    '''
    Arbitrary function annotated *only* by a single keyword-only parameter.
    Yes, this previously blew up. It's best *not* to ask why. We shudder.
    '''

    return wrinkled_his + quivering_lips


def for_sleep(he_knew, /, kept_most_relentlessly: str):
    '''
    Arbitrary function annotated *only* by a single positional-only parameter.
    Yes, this previously blew up. It's best *not* to ask why. We shudder.
    '''

    return he_knew + kept_most_relentlessly


# Assert these functions return the expected values when passed the expected
# parameters.
assert of_desperate_hope(
    wrinkled_his='Of desperate hope ',
    quivering_lips='wrinkled his quivering lips.',
) == 'Of desperate hope wrinkled his quivering lips.'
assert for_sleep('For sleep, he knew, ', 'kept most relentlessly') == (
    'For sleep, he knew, kept most relentlessly')

# Assert these functions raise the expected exceptions when passed unexpected
# parameters.
with raises(BeartypeCallHintParamViolation):
    of_desperate_hope(
        wrinkled_his=b'Its precious charge, ',
        quivering_lips=b'and silent death exposed,',
    )
with raises(BeartypeCallHintParamViolation):
    for_sleep(b'Faithless perhaps as sleep, ', b'a shadowy lure,')

# ....................{ CLASSES                            }....................
class ConfessStAugust(object):
    '''
    Arbitrary class to be implicitly decorated by the :func:`beartype.beartype`
    decorator by the :func:`beartype.claw.beartype_this_package` import hook
    installed by the parent ``beartype_this_package.__init__`` submodule.
    '''

    def with_aught_of_natural_piety(self, to_feel: Union[complex, str]) -> (
        Union[List[bytes], complex]):
        '''
        Arbitrary method accepting the passed object under the non-default
        beartype configuration established by the sibling ``__init__`` submodule
        in this subpackage and returning that object as is, enabling callers to
        trivially test whether any call to this method violates the type hints
        annotating this method.
        '''

        # Look, @beartype. Just do it!
        return to_feel


    @beartype  # <-- prefer the default beartype configuration
    def your_love(
        self, and_recompense_the_boon_with_mine: Union[complex, bool]) -> (
        Union[complex, List[bytes]]):
        '''
        Arbitrary method accepting the passed object under the default beartype
        configuration and returning that object as is, enabling callers to
        trivially test whether any call to this method violates the type hints
        annotating this method.
        '''

        # Look here, you. Just do it yet again!
        return and_recompense_the_boon_with_mine


# Arbitrary instance of this class.
if_dewy_morn = ConfessStAugust()

# Assert that calling the first method passed an arbitrary integer returns that
# integer as is *WITHOUT* raising an exception.
assert if_dewy_morn.with_aught_of_natural_piety(len(
    'If dewy morn, and odorous noon, and even')) == 40

# Assert that calling the first method passed an invalid parameter raises the
# expected exception.
with raises(BeartypeCallHintParamViolation):
    if_dewy_morn.with_aught_of_natural_piety(
        b'With sunset and its gorgeous ministers,')

# Assert that calling the second method passed an arbitrary complex number
# returns that complex number as is *WITHOUT* raising an exception.
assert if_dewy_morn.your_love(len(
    "And solemn midnight's tingling silentness;") + 1j) == 42 + 1j

# Assert that calling the second method passed an invalid parameter raises the
# expected exception.
with raises(BeartypeCallHintParamViolation):
    if_dewy_morn.your_love(len("If autumn's hollow sighs in the sere wood,"))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
:mod:`pytest` **PEP-agnostic type hint utilities.**
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from pytest import fixture

# ....................{ CLASSES                            }....................
class HintPithMetadata(object):
    '''
    Dataclass encapsulating all relevant type hint- and pith-specific metadata
    iteratively yielded by each iteration of the :func:`.iter_hints_piths_meta`
    generator.

    Attributes
    ----------
    hint_meta : HintNonpepMetadata
        Metadata describing the currently iterated type hint.
    pith_meta : HintPithSatisfiedMetadata
        Metadata describing this ``pith``.
    pith : object
        Object either satisfying or violating this hint.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,
        hint_meta: 'HintNonpepMetadata',
        pith_meta: 'HintPithSatisfiedMetadata',
        pith: object,
    ) -> None:
        '''
        Initialize this dataclass.

        Parameters
        ----------
        hint_meta : HintNonpepMetadata
            Metadata describing the currently iterated type hint.
        pith_meta : Union[HintPithSatisfiedMetadata, HintPithUnsatisfiedMetadata]
            Metadata describing this ``pith``.
        pith : object
            Object either satisfying or violating this hint.
        '''

        # Classify all passed parameters. For simplicity, avoid validating
        # these parameters; we simply *CANNOT* be bothered at the moment.
        self.hint_meta = hint_meta
        self.pith_meta = pith_meta
        self.pith = pith

    # ..................{ DUNDERS                            }..................
    def __repr__(self) -> str:
        return '\n'.join((
            f'{self.__class__.__name__}(',
            f'    hint_meta={repr(self.hint_meta)},',
            f'    pith_meta={repr(self.pith_meta)},',
            f'    pith={repr(self.pith)},',
            f')',
        ))

# ....................{ ITERATORS                          }....................
@fixture(scope='session')
def iter_hints_piths_meta(hints_meta) -> (
    'Callable[[], Iterable[HintPithMetadata]]'):
    '''
    Session-scoped fixture yielding a **PEP-agnostic type hint-pith metadata
    generator** (i.e., factory function creating and returning a generator
    iteratively yielding :class:`.HintPithMetadata` instances, each describing a
    sample type hint exercising an edge case in the :mod:`beartype` codebase
    paired with a related object either satisfying or violating this hint --
    including both PEP-compliant and -noncompliant type hints).

    Caveats
    -------
    This fixture *cannot* be directly iterated. Instead, this fixture must be
    first called and then iterated by callers. Why? Because
    :mod:`pytest.fixture`. By design, pytest ignores all ``yield`` statements in
    a fixture except the first. Ergo, the only means of defining a fixture
    generator is to define a fixture returning a generator closure. See also
    `this relevant StackOverflow answer <answer_>`__.

    .. _answer:
       https://stackoverflow.com/a/66397211/2809027

    Parameters
    ----------
    hints_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintNonpepMetadata]
        List of PEP-agnostic type hint metadata describing sample PEP-agnostic
        type hints exercising edge cases in the :mod:`beartype` codebase.

    Yields
    ------
    Callable[[], Iterable[HintPithMetadata]]
        PEP-agnostic type hint-pith metadata generator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.utilobject import is_object_context_manager
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPithSatisfiedMetadata,
        HintPithUnsatisfiedMetadata,
    )
    from beartype_test._util.pytcontext import noop_context_manager

    # Tuple of two arbitrary values used to trivially iterate twice below.
    RANGE_2 = (None, None,)

    # ....................{ CLOSURE                        }....................
    def hints_piths_meta() -> 'Iterable[HintPithMetadata]':
        '''
        **PEP-agnostic type hint-pith metadata generator** (i.e., factory
        function creating and returning a generator iteratively yielding
        :class:`.HintPithMetadata` instances, each describing a sample type hint
        exercising an edge case in the :mod:`beartype` codebase paired with a
        related object either satisfying or violating this hint -- including
        both PEP-compliant and -noncompliant type hints).
        '''

        # For each predefined type hint and associated metadata...
        for hint_meta in hints_meta:
            # print(f'Type-checking type hint {repr(hint_meta.hint)}...')

            # If this hint is currently unsupported, continue to the next.
            if not hint_meta.is_supported:
                continue
            # Else, this hint is currently supported.

            # Repeat the following logic twice. Why? To exercise memoization
            # across repeated @beartype decorations on different callables
            # annotated by the same hints.
            for _ in RANGE_2:
                # For each pith either satisfying or violating this hint...
                for pith_meta in hint_meta.piths_meta:
                    # Assert this metadata is an instance of the desired dataclass.
                    assert isinstance(pith_meta, HintPithSatisfiedMetadata)

                    # Pith to be type-checked against this hint, defined as...
                    pith = (
                        # If this pith is actually a pith factory (i.e., callable
                        # accepting *NO* parameters and dynamically creating and
                        # returning the value to be used as the desired pith), call
                        # this factory and localize its return value.
                        pith_meta.pith()
                        if pith_meta.is_pith_factory else
                        # Else, localize this pith as is.
                        pith_meta.pith
                    )
                    # print(f'Type-checking PEP type hint {repr(hint_meta.hint)} against {repr(pith)}...')

                    # Context manager under which to validate this pith against
                    # this hint, defined as either...
                    pith_context_manager = (
                        # This pith itself if both...
                        pith
                        if (
                            # This pith is a context manager *AND*...
                            is_object_context_manager(pith) and
                            # This pith should be safely opened and closed as a
                            # context rather than preserved as a context manager...
                            not pith_meta.is_context_manager
                        ) else
                        # Else, the noop context manager yielding this pith.
                        noop_context_manager(pith)
                    )

                    # With this pith safely opened and closed as a context...
                    with pith_context_manager as pith_context:
                        # If this pith does *NOT* satisfy this hint...
                        if isinstance(pith_meta, HintPithUnsatisfiedMetadata):
                            # Assert that iterables of uncompiled regular
                            # expression expected to match and *NOT* match this
                            # message are *NOT* strings, as commonly occurs when
                            # accidentally omitting a trailing comma in tuples
                            # containing only one string: e.g.,
                            # * "('This is a tuple, yo.',)" is a 1-tuple
                            #   containing one string.
                            # * "('This is a string, bro.')" is a string *NOT*
                            #   contained in a 1-tuple.
                            assert not isinstance(
                                pith_meta.exception_str_match_regexes, str)
                            assert not isinstance(
                                pith_meta.exception_str_not_match_regexes, str)

                        # Yield this metadata to the caller.
                        yield HintPithMetadata(
                            hint_meta=hint_meta,
                            pith_meta=pith_meta,
                            pith=pith_context,
                        )

    # ....................{ YIELD                          }....................
    # Yield this closure.
    yield hints_piths_meta



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`604`-compliant **union type hint utilities**
(i.e., low-level callables handling union type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep604Exception
from beartype._cave._cavefast import (
    HintPep604Type,
    HintPep604ItemTypes,
)
from beartype._data.kind.datakindiota import SENTINEL
from beartype._data.typing.datatypingport import (
    Hint,
    SequenceHints,
    TypeIs,
)
from beartype._util.cache.utilcachecall import callable_cached
from collections.abc import Sequence
from functools import reduce
from operator import __or__ as or_operator

# ....................{ RAISERS                        }....................
#FIXME: Unit test us up, please.
def die_if_hint_pep604_inconsistent(hint: Hint) -> None:
    '''
    Raise an exception if the passed object is a :pep:`604`-compliant
    **inconsistent type hint** (i.e., object permissible as an item of a
    :pep:`604`-compliant new union whose machine-readable representation is
    *not* the machine-readable representation of this hint in new unions).

    Motivation
    ----------
    This raiser protects the :mod:`beartype` codebase against inconsistencies in
    poorly implemented third-party type hints whose **machine-readable
    representations** (i.e., strings returned by passing those hints to the
    :func:`repr` builtin) differ from their representations in new unions
    containing those hints.

    Ideally, *no* such inconsistencies would ever exist. For unknown reasons,
    some third-party type hints induce such inconsistencies. :mod:`nptyping` is
    the canonical example. Type hint factories published by :mod:`nptyping`
    dynamically create in-memory classes all sharing the same fully-qualified
    names despite being distinct classes. Ye who code, grok and weep: e.g.,

    .. code-block:: pycon

       # Define two typical "nptyping" type hints.
       >>> from nptyping import Float64, NDArray, Shape
       >>> foo = NDArray[Shape["N, N"], Float64]
       >>> bar = NDArray[Shape["*"], Float64]

       >>> repr(foo)
       NDArray[Shape["N, N"], Float64]  # <-- this is sane
       >>> repr(bar)
       NDArray[Shape["*"], Float64]  # <-- this is sane, too

       >>> foo == bar
       False    # <-- still sane
       >>> foo.__name__
       NDArray  # <-- this is insane
       >>> bar.__name__
       NDArray  # <-- still insane after all these years

       >>> foo | None == bar | None
       False           # <-- back to sane
       >>> repr(foo | None)
       NDArray | None  # <-- big yikes
       >>> repr(bar | None)
       NDArray | None  # <-- yikes intensifies

    ``foo`` and ``bar`` are distinct type hints matching different NumPy arrays;
    their representations are likewise distinct. The new unions of those hints
    with :data:`None` are also distinct type hints; nonetheless, the
    representations of those new unions **are the exact same.**

    This raiser detects this inconsistency by raising an exception from the
    :func:`beartype.beartype` decorator. If we failed to do so, then
    :func:`beartype.beartype` would behave non-deterministically when presented
    with such hints. Consider the following decoration:

    .. code-block:: python

       @beartype
       def bad_func(first_array: foo | None, second_array: bar | None) -> None:
           ...

    Given that decoration, :func:`beartype.beartype` would (in order):

    #. Cache the first new union ``foo | None`` under the string representation
       ``"NDArray | None"``.
    #. Erroneously replace the second new union ``bar | None`` with the
       previously cached new union ``foo | None``. Why? Because those two new
       unions share the same representations. From the limited perspective of
       :mod:`beartype`, those two new unions are effectively the same new union
       and thus can be safely de-duplicated. *This is why we facepalm.*

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Raises
    ------
    BeartypeDecorHintPep604Exception
        If this object is a :pep:`604`-compliant inconsistent type hint.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.utilhintget import get_hint_repr

    # If this hint is invalid as an item of a PEP 604-compliant new union,
    # silently reduce to a noop.
    if not isinstance(hint, HintPep604ItemTypes):
        return
    # Else, this hint is valid as an item of a PEP 604-compliant new union.

    # Machine-readable representation of this hint.
    hint_repr = get_hint_repr(hint)

    # If this representation is prefixed by the "<" character, this
    # representation is assumed to be (at least *SOMEWHAT*) standardized and
    # thus internally consistent. This includes:
    # * Standard classes (e.g., "<class 'bool'>").
    # * @beartype-specific forward reference subclasses (e.g., "<forwardref
    #   UndeclaredClass(__beartype_scope__='some_package')>").
    #
    # This is *NOT* simply an optimization. Standardized representations
    # *MUST* be excluded from consideration, as the representations of new
    # unions containing these hints is *NOT* prefixed by "<": e.g.,
    #     >>> repr(bool)
    #     <class 'bool'>
    #     >>> bool | None
    #     bool | None
    if hint_repr[0] == '<':
        return
    # Else, this representation is *NOT* prefixed by the "<" character.

    # Arbitrary PEP 604-compliant new union defined as the conjunction
    # of this hint with an arbitrary builtin type guaranteed to exist.
    #
    # Note that order is significant.
    hint_pep604 = hint | int  # type: ignore[operator]

    # Machine-readable representation of this new union.
    hint_pep604_repr = get_hint_repr(hint_pep604)

    # If the representation of this new union is *NOT* prefixed by the
    # representation of this hint, raise an exception.
    if not hint_pep604_repr.startswith(hint_repr):
        raise BeartypeDecorHintPep604Exception(
            f'Type hint {hint_repr} inconsistent with respect to '
            f'repr() strings: e.g.,\n'
            f'\t>>> repr({hint_repr})\n'
            f'\t{hint_repr}  # <-- this is fine\n'
            f'\t>>> repr({hint_repr} | int)\n'
            f'\t{hint_pep604_repr}  # <-- *THIS IS REALLY SUPER BAD*\n'
            f'\n'
            f'\t# Ideally, that output should instead resemble:\n'
            f'\t>>> repr({hint_repr} | int)\n'
            f'\t{hint_repr} | int  # <-- what @beartype wants!\n'
            f'\n'
            f'Inconsistent type hints are unsupported by @beartype, '
            f'which requires consistency between type hints and '
            f'repr() strings. Consider reporting this issue to the '
            f'third-party developers implementing this hint, which '
            f'(probably) fails to define the PEP 585-compliant '
            f'"__args__" and "__origin__" dunder attributes standardized '
            f'by the "types.GenericAlias" API. See also:\n'
            f'\thttps://peps.python.org/pep-0585/#parameters-to-generics-are-available-at-runtime'
        )
    # Else, the representation of this new union is prefixed by the
    # representation of this hint as expected.

# ....................{ TESTERS                        }....................
def is_hint_pep604(hint: object) -> TypeIs[Hint]:
    '''
    :data:`True` only if the passed object is a :pep:`604`-compliant **union**
    (i.e., ``|``-delimited disjunction of two or more isinstanceable types).

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a :pep:`604`-compliant union.
    '''

    # Release the werecars, Bender!
    return isinstance(hint, HintPep604Type)


# ....................{ FACTORIES                          }....................
#FIXME: Unit test us up, please.
@callable_cached
def make_hint_pep484604_union(hint_childs: SequenceHints) -> Hint:
    '''
    :pep:`604`- or :pep:`484`-compliant union type hint synthesized from the
    passed sequence of two or more PEP-compliant type hints if this sequence
    contains two or more items, the one PEP-compliant type hint in this sequence
    if this sequence contains only one item, *or* raise an exception otherwise
    (i.e., if this sequence is empty).

    This factory preferentially creates and returns (in descending order of
    preference):

    * If *all* of the items in the passed sequence are valid items of a
      :pep:`604`-compliant **new-style union** (``|``-delimited union of two or
      more types under Python >= 3.10), that union of these items.
    * Else, a :pep:`484`-compliant **old-style union** (i.e.,
      :obj:`typing.Union` subscription) of these items.

    This factory is memoized for efficiency. Technically, the
    :attr:`typing.Union` type hint factory already caches its subscripted
    arguments. Pragmatically, that caching is slow and thus worth optimizing
    with trivial optimization on our end. Moreover, this factory is called by
    the performance-sensitive
    :func:`beartype._check.convert._convcoerce.coerce_hint_any` coercer in an
    early-time code path of the :func:`beartype.beartype` decorator. Optimizing
    this factory thus optimizes :func:`beartype.beartype` itself. Even more
    importantly, :pep:`604`-compliant new-style unions are *not* self-caching:

    .. code-block:: pycon

       >>> int | bool is int | bool
       False  # <-- *BIG YIKES*

    Parameters
    ----------
    hint_childs : SequenceHints
        Sequence of all child type hints to be unioned together.

    Returns
    -------
    Hint
        Either:

        * If this sequence contains two or more items, the union type hint
          synthesized from these items.
        * If this sequence contains only one item, this item as is.

    Raises
    ------
    BeartypeDecorHintPep604Exception
        If this sequence is empty.
    '''
    assert isinstance(hint_childs, Sequence), (
        f'{repr(hint_childs)} not sequence.')

    # If this sequence is empty, raise an exception.
    if not hint_childs:
        raise BeartypeDecorHintPep604Exception('"hints" sequence empty.')
    # Else, this sequence contains one or more child type hints.
    #
    # If this sequence contains only one child type hint, return this hint
    # unmodified as is.
    elif len(hint_childs) == 1:
        return hint_childs[0]
    # Else, this sequence contains two or more child type hints.
    #
    # If this sequence is *NOT* already a tuple, coerce this sequence into a
    # tuple. The lower-level functions called below all expected a tuple, oddly.
    elif not isinstance(hint_childs, tuple):
        hint_childs = tuple(hint_childs)
    # Else, this sequence is already a tuple. In this case, preserve this tuple.

    # Union of these child type hints to be returned.
    hint_union: Hint = SENTINEL  # type: ignore[assignment]

    # Attempt to dynamically fabricate a PEP 604-compliant new-style union of
    # these items if these items are all PEP 604-compliant.
    try:
        hint_union = reduce(or_operator, hint_childs)  # type: ignore[assignment]
    # If *ANY* exception whatsoever is raised, one or more of these items are
    # PEP 604-noncompliant. In this case, silently ignore this exception in
    # favour of falling back to a PEP 484-compliant old-style union below. We
    # don't make breaky. We only fix breaky, people.
    except Exception:
        pass

    # If a PEP 604-compliant new-style union was *NOT* fabricated above...
    if hint_union is SENTINEL:
        # Avoid circular import dependencies.
        from beartype._util.hint.pep.proposal.pep484.pep484union import (
            make_hint_pep484_union)

        # Fallback to a PEP 484-compliant old-style union.
        hint_union = make_hint_pep484_union(hint_childs)
    # Else, a PEP 604-compliant new-style union was fabricated above.

    # Return this union.
    return hint_union



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **beartype import hook data type subpackage initialization submodule**
(i.e., data module mimicking real-world usage of various :func:`beartype.claw`
import hooks on various types of objects, including both callables and
non-callable containers).
'''

# ....................{ IMPORTS                            }....................
from beartype_test.a00_unit.data.claw.intraprocess.hookable_package.kind import (
    # Intentionally exercise synchronous functions first. As the simplest form
    # of decoratable object supported by @beartype, synchronous functions form
    # the core backbone of @beartype. If @beartype fails to decorate even
    # synchronous functions, all hope is lost for less trivial decoration below.
    data_claw_func,

    # Intentionally exercise coroutines (i.e., asynchronous functions) next.
    # Although still simple, coroutines are slightly less trivial than
    # synchronous functions for @beartype to decorate.
    data_claw_coro,

    # Intentionally exercise types (i.e., user-defined classes) next. Types are
    # considerably higher-level and thus less trivial than either synchronous
    # or asynchronous functions for @beartype to decorate.
    data_claw_type,
)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **import hook path hook registrars** (i.e., high-level functions both
adding and removing our beartype import path hook singleton to and from the
front of the standard :mod:`sys.path_hooks` list, which when added recursively
applies the :func:`beartype.beartype` decorator to all well-typed callables and
classes defined by all submodules of all packages previously registered by a
call to a public :func:`beartype.claw` import hook).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.claw._importlib._clawimpload import BeartypeSourceFileLoader
from importlib import invalidate_caches
from importlib.machinery import (
    BYTECODE_SUFFIXES,
    SOURCE_SUFFIXES,
    FileFinder,
    ExtensionFileLoader,
    SourcelessFileLoader,
)
from sys import (
    path_hooks,
    path_importer_cache,
)

# Intentionally violate privacy encapsulate in the standard Python library,
# because there is *NO* valid alternative. This low-level private getter
# function returns a tuple of the filetypes of *ALL* C extensions supported by
# the current platform (e.g., as shared libraries).
from _imp import extension_suffixes

# ....................{ ADDERS                             }....................
#FIXME: Unit test us up, please.
def add_beartype_pathhook() -> None:
    '''
    Add our **beartype import path hook singleton** (i.e., single callable
    guaranteed to be inserted at most once to the front of the standard
    :mod:`sys.path_hooks` list recursively applying the
    :func:`beartype.beartype` decorator to all well-typed callables and classes
    defined by all submodules of all packages previously registered by a call to
    a public :func:`beartype.claw` function) if this path hook has yet to be
    added *or* silently reduce to a noop otherwise (i.e., if this path hook has
    already been added).

    Caveats
    -------
    **This function is non-thread-safe.** For both simplicity and efficiency,
    the caller is expected to provide thread-safety through a higher-level
    locking primitive managed by the caller.

    See Also
    --------
    https://stackoverflow.com/a/43573798/2809027
        StackOverflow answer strongly inspiring the low-level implementation of
        this function with respect to inscrutable :mod:`importlib` machinery.
    '''

    # Avoid circular import dependencies.
    from beartype.claw._clawstate import claw_state

    # If this function has already been called under the active Python
    # interpreter, silently reduce to a noop.
    if claw_state.beartype_pathhook is not None:
        return
    # Else, this function has *NOT* yet been called under this interpreter.

    # Closure instantiating a new "FileFinder" instance invoking this loader.
    #
    # Note that we intentionally ignore mypy complaints here. Why? Because mypy
    # erroneously believes this method accepts 2-tuples whose first items are
    # loader *INSTANCES* (e.g., "Tuple[Loader, List[str]]"). In fact, this
    # method accepts 2-tuples whose first items are loader *TYPES* (e.g.,
    # "Tuple[Type[Loader], List[str]]"). This is why we can't have nice things.
    loader_factory = FileFinder.path_hook(*_LOADERS_DETAILS)  # type: ignore[arg-type]

    # Prepend a new path hook (i.e., factory closure encapsulating this loader)
    # *BEFORE* all other path hooks.
    path_hooks.insert(0, loader_factory)
    # path_hooks.append(loader_factory)

    #FIXME: Uncomment as needed to debug the contents of the "path_hooks" list.
    # print(f'path_hooks: {path_hooks}')
    # for path_hook in path_hooks:
    #     try:
    #         file_finder = path_hook('/usr/lib/python3.11')
    #         print(f'file_finder: {file_finder} [{file_finder._loaders}]')
    #     except Exception:
    #         pass

    # Prevent subsequent calls to this function from erroneously re-adding
    # duplicate copies of this path hook immediately *AFTER* successfully adding
    # the first such path hook.
    #
    # Note that we intentionally avoid globalizing this path hook until *AFTER*
    # successfully having done so. Why? Negligible safety. The companion
    # remove_beartype_pathhook() function raises a non-human-readable exception
    # if this global is non-"None" but *NOT* in the "path_hooks" list.
    claw_state.beartype_pathhook = loader_factory

    # Lastly, clear *ALL* import path hook caches for safety.
    _clear_importlib_caches()

# ....................{ REMOVERS                           }....................
#FIXME: Unit test us up, please.
def remove_beartype_pathhook() -> None:
    '''
    Remove our **beartype import path hook singleton** (i.e., single callable
    guaranteed to be inserted at most once to the front of the standard
    :mod:`sys.path_hooks` list recursively applying the
    :func:`beartype.beartype` decorator to all well-typed callables and classes
    defined by all submodules of all packages previously registered by a call to
    a public :func:`beartype.claw` function) if this path hook has already been
    added *or* silently reduce to a noop otherwise (i.e., if this path hook has
    yet to be added).

    Caveats
    -------
    **This function is non-thread-safe.** For both simplicity and efficiency,
    the caller is expected to provide thread-safety through a higher-level
    locking primitive managed by the caller.
    '''

    # Avoid circular import dependencies.
    from beartype.claw._clawstate import claw_state

    # If the add_beartype_pathhook() function has *NOT* yet been called under
    # the active Python interpreter, silently reduce to a noop.
    if claw_state.beartype_pathhook is None:
        return
    # Else, that function has already been called under this interpreter.

    # Remove the prior path hook added by that function *OR* raise a
    # non-human-readable "ValueError" exception if this global is non-"None" but
    # *NOT* in the "path_hooks" list (which should *NEVER* happen, but it will).
    path_hooks.remove(claw_state.beartype_pathhook)

    # Allow subsequent calls to the add_beartype_pathhook() to re-add a new
    # instance of this path hook immediately *AFTER* successfully removing the
    # first such path hook.
    claw_state.beartype_pathhook = None

    # Lastly, clear *ALL* import path hook caches for safety.
    _clear_importlib_caches()

# ....................{ PRIVATE ~ globals                  }....................
_LOADERS_DETAILS = (
    # Beartype-agnostic C extension loader details. Since C extensions *CANNOT*
    # (by definition) be decompiled into an abstract syntax tree (AST), beartype
    # has *NO* means of decorating C extensions. Ergo, we necessarily defer to
    # Python's default C extension loader.
    (ExtensionFileLoader, extension_suffixes()),

    # Beartype-specific **source module loader** (i.e., file loader loading
    # uncompiled pure-Python modules of the filetype ".py").
    (BeartypeSourceFileLoader, SOURCE_SUFFIXES),

    #FIXME: Generalize this into a beartype-specific bytecode module loader.
    #How? By leveraging the third-party "astor" package, which provides a
    #code_to_ast() function decompiling arbitrary code objects into ASTs. See
    #also this relevant StackOverflow answer by myself:
    #    https://stackoverflow.com/a/76641537/2809027

    # Beartype-agnostic **bytecode module loader** (i.e., file loader loading
    # precompiled pure-Python modules from bytecode files compiled in
    # "__pycache__/" subdirectories that lack corresponding uncompiled
    # pure-Python modules of the filetype ".py").
    (SourcelessFileLoader, BYTECODE_SUFFIXES),
)
'''
Tuple of all **file-based module loader details** (i.e., 2-tuple ``(file_loader,
filetypes)`` of the undocumented format expected by the
:meth:`FileFinder.path_hook` class method called by the
:func:`beartype.claw._importlib.clawimpmain.add_beartype_pathhook` function,
associating each file-based module loader with the platform-specific filetypes
of all modules loaded by that module).

We didn't do it. Don't blame the bear.

See Also
--------
:func:`importlib.machinery._get_supported_file_loaders`
    Low-level private getter function strongly inspiring the definition of this
    global, which implements nearly identical functionality (albeit in a
    :mod:`beartype`-specific manner).
'''

# ....................{ PRIVATE ~ cachers                  }....................
#FIXME: Unit test us up, please.
def _clear_importlib_caches() -> None:
    '''
    Clear *all* :mod:`sys`- and :mod:`importlib`-specific caches pertaining to
    **import path hooks** (i.e., the standard :mod:`sys.path_hooks` list).

    This function is typically called immediately *after* our beartype import
    path hook singleton is either added to or removed from the path hooks list.
    '''

    # Uncache *ALL* competing loaders cached by prior importations. Just do it!
    path_importer_cache.clear()

    # Clear *ALL* "importlib" caches as well for safety.
    invalidate_caches()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`586` **type-checking expression snippets** (i.e., triple-quoted
pure-Python string constants formatted and concatenated together to dynamically
generate boolean expressions type-checking arbitrary objects against
:pep:`586`-compliant type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.typing.datatyping import CallableStrFormat

# ....................{ CODE                               }....................
CODE_PEP586_PREFIX = '''(
{{indent_curr}}    # True only if this pith is of one of these literal types.
{{indent_curr}}    isinstance({pith_curr_assign_expr}, {hint_child_types_expr}) and ('''
'''
:pep:`586`-compliant code snippet prefixing all code type-checking the current
pith against a :pep:`586`-compliant :class:`typing.Literal` type hint
subscripted by one or more literal objects.
'''


CODE_PEP586_SUFFIX = '''
{indent_curr}))'''
'''
:pep:`586`-compliant code snippet suffixing all code type-checking the current
pith against a :pep:`586`-compliant :class:`typing.Literal` type hint
subscripted by one or more literal objects.
'''


CODE_PEP586_LITERAL = '''
{{indent_curr}}        # True only if this pith is equal to this literal.
{{indent_curr}}        {pith_curr_var_name} == {hint_child_expr} or'''
'''
:pep:`586`-compliant code snippet type-checking the current pith against the
current child literal object subscripting a :pep:`586`-compliant
:class:`typing.Literal` type hint.

Caveats
-------
The caller is required to manually slice the trailing suffix ``" and"`` after
applying this snippet to the last subscripted argument of such a
:class:`typing.Literal` type. While there exist alternate and more readable
means of accomplishing this, this approach is the optimally efficient.

The ``{indent_curr}`` format variable is intentionally brace-protected to
efficiently defer its interpolation until the complete PEP-compliant code
snippet type-checking the current pith against *all* subscripted arguments of
this parent hint has been generated.
'''

# ....................{ FORMATTERS                         }....................
# str.format() methods, globalized to avoid inefficient dot lookups elsewhere.
# This is an absurd micro-optimization. *fight me, github developer community*
CODE_PEP586_LITERAL_format: CallableStrFormat = CODE_PEP586_LITERAL.format
CODE_PEP586_PREFIX_format: CallableStrFormat = CODE_PEP586_PREFIX.format



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype cave.**

This submodule collects common types (e.g., :class:`NoneType`, the type of the
``None`` singleton) and tuples of common types (e.g., :data:`CallableTypes`, a
tuple of the types of all callable objects).

PEP 484
----------
This module is intentionally *not* compliant with the :pep:`484` standard
implemented by the stdlib :mod:`typing` module, which formalizes type hinting
annotations with a catalogue of generic classes and metaclasses applicable to
common use cases. :mod:`typing` enables end users to enforce contractual
guarantees over the contents of arbitrarily complex data structures with the
assistance of third-party static type checkers (e.g., :mod:`mypy`,
:mod:`pyre`), runtime type checkers (e.g., :mod:`beartype`, :mod:`typeguard`),
and integrated development environments (e.g., PyCharm).

Genericity comes at a cost, though. Deeply type checking a container containing
``n`` items, for example, requires type checking both that container itself
non-recursively *and* each item in that container recursively. Doing so has
time complexity ``O(N)`` for ``N >= n`` the total number of items transitively
contained in this container (i.e., items directly contained in this container
*and* items directly contained in containers contained in this container).
While the cost of this operation can be paid either statically *or* amortized
at runtime over all calls to annotated callables accepting that container, the
underlying cost itself remains the same.

By compare, this module only contains standard Python classes and tuples of
such classes intended to be passed as is to the C-based :func:`isinstance`
builtin and APIs expressed in terms of that builtin (e.g., :mod:`beartype`).
This module only enables end users to enforce contractual guarantees over the
types but *not* contents of arbitrarily complex data structures. This
intentional tradeoff maximizes runtime performance at a cost of ignoring the
types of items contained in containers.

In summary:

=====================  ====================  ====================================
feature set            :mod:`beartype.cave`  :mod:`typing`
=====================  ====================  ====================================
type checking          **shallow**           **deep**
type check items?      **no**                **yes**
:pep:`484`-compliant?  **no**                **yes**
time complexity        ``O(1)``              ``O(N)``
performance            stupid fast           *much* less stupid fast
implementation         C-based builtin call  pure-Python (meta)class method calls
low-level primitive    :func:`isinstance`    :mod:`typing.TypingMeta`
=====================  ====================  ====================================
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: *NEVER IMPORT FROM THIS SUBPACKAGE FROM WITHIN BEARTYPE ITSELF.*
# This subpackage currently imports from expensive third-party packages on
# importation (e.g., NumPy) despite beartype itself *NEVER* requiring those
# imports. Until resolved, this subpackage is considered tainted.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To prevent "mypy --no-implicit-reexport" from raising literally
# hundreds of errors at static analysis time, *ALL* public attributes *MUST* be
# explicitly reimported under the same names with "{exception_name} as
# {exception_name}" syntax rather than merely "{exception_name}". Yes, this is
# ludicrous. Yes, this is mypy. For posterity, these failures resemble:
#     beartype/_cave/_cavefast.py:47: error: Module "beartype.roar" does not
#     explicitly export attribute "BeartypeCallUnavailableTypeException";
#     implicit reexport disabled  [attr-defined]
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid polluting the public module namespace, external attributes
# should be locally imported at module scope *ONLY* under alternate private
# names (e.g., "from argparse import ArgumentParser as _ArgumentParser" rather
# than merely "from argparse import ArgumentParser").
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from beartype.cave._cavelib import (
    # Types.
    ArgParserType as ArgParserType,
    ArgSubparsersType as ArgSubparsersType,
    WeakRefCType as WeakRefCType,

    # Type tuples.
    WeakRefProxyCTypes as WeakRefProxyCTypes,
)
from beartype._cave._caveabc import (
    BoolType as BoolType,
)
from beartype._cave._cavefast import (
    # Types.
    AnyType as AnyType,
    AsyncCoroutineCType as AsyncCoroutineCType,
    AsyncGeneratorCType as AsyncGeneratorCType,
    CallableCodeObjectType as CallableCodeObjectType,
    ClassDictType as ClassDictType,
    CallableFrameType as CallableFrameType,
    CallableFunctoolsPartialType as CallableFunctoolsPartialType,
    ClassType as ClassType,
    ClosureVarCellType as ClosureVarCellType,
    CollectionType as CollectionType,
    ContainerType as ContainerType,
    EllipsisType as EllipsisType,
    EnumType as EnumType,
    EnumMemberType as EnumMemberType,
    ExceptionTracebackType as ExceptionTracebackType,
    FileType as FileType,
    FunctionType as FunctionType,
    FunctionOrMethodCType as FunctionOrMethodCType,
    GeneratorCType as GeneratorCType,
    GeneratorType as GeneratorType,
    HashableType as HashableType,
    HintGenericSubscriptedType as HintGenericSubscriptedType,
    IntOrFloatType as IntOrFloatType,
    IntType as IntType,
    IterableType as IterableType,
    IteratorType as IteratorType,
    MappingMutableType as MappingMutableType,
    MappingType as MappingType,
    MethodBoundInstanceDunderCType as MethodBoundInstanceDunderCType,
    MethodBoundInstanceOrClassType as MethodBoundInstanceOrClassType,
    MethodDecoratorClassType as MethodDecoratorClassType,
    MethodDecoratorPropertyType as MethodDecoratorPropertyType,
    MethodDecoratorStaticType as MethodDecoratorStaticType,
    MethodUnboundClassCType as MethodUnboundClassCType,
    MethodUnboundInstanceDunderCType as MethodUnboundInstanceDunderCType,
    MethodUnboundInstanceNondunderCType as MethodUnboundInstanceNondunderCType,
    MethodUnboundPropertyNontrivialCExtensionType as
        MethodUnboundPropertyNontrivialCExtensionType,
    MethodUnboundPropertyTrivialCExtensionType as
        MethodUnboundPropertyTrivialCExtensionType,
    ModuleType as ModuleType,
    NoneType as NoneType,
    NotImplementedType as NotImplementedType,
    NumberRealType as NumberRealType,
    NumberType as NumberType,
    SizedType as SizedType,
    QueueType as QueueType,
    RegexCompiledType as RegexCompiledType,
    RegexMatchType as RegexMatchType,
    SetType as SetType,
    SequenceMutableType as SequenceMutableType,
    SequenceType as SequenceType,
    StrType as StrType,
    UnavailableType as UnavailableType,

    # Type tuples.
    AsyncCTypes as AsyncCTypes,
    BoolOrNumberTypes as BoolOrNumberTypes,
    CallableCTypes as CallableCTypes,
    CallableOrClassTypes as CallableOrClassTypes,
    CallableOrStrTypes as CallableOrStrTypes,
    CallableTypes as CallableTypes,
    DecoratorTypes as DecoratorTypes,
    FunctionTypes as FunctionTypes,
    ModuleOrStrTypes as ModuleOrStrTypes,
    MethodBoundTypes as MethodBoundTypes,
    MethodDecoratorBuiltinTypes as MethodDecoratorBuiltinTypes,
    MethodUnboundTypes as MethodUnboundTypes,
    MethodTypes as MethodTypes,
    MappingOrSequenceTypes as MappingOrSequenceTypes,
    ModuleOrSequenceTypes as ModuleOrSequenceTypes,
    NumberOrIterableTypes as NumberOrIterableTypes,
    NumberOrSequenceTypes as NumberOrSequenceTypes,
    RegexTypes as RegexTypes,
    ScalarTypes as ScalarTypes,
    TestableTypes as TestableTypes,
    UnavailableTypes as UnavailableTypes,
)
from beartype._cave._cavemap import (
    NoneTypeOr as NoneTypeOr,
)

# ....................{ DEPRECATIONS                       }....................
def __getattr__(attr_deprecated_name: str) -> object:
    '''
    Dynamically retrieve a deprecated attribute with the passed unqualified
    name from this submodule and emit a non-fatal deprecation warning on each
    such retrieval if this submodule defines this attribute *or* raise an
    exception otherwise.

    The Python interpreter implicitly calls this :pep:`562`-compliant module
    dunder function under Python >= 3.7 *after* failing to directly retrieve an
    explicit attribute with this name from this submodule. Since this dunder
    function is only called in the event of an error, neither space nor time
    efficiency are a concern here.

    Parameters
    ----------
    attr_deprecated_name : str
        Unqualified name of the deprecated attribute to be retrieved.

    Returns
    ----------
    object
        Value of this deprecated attribute.

    Warns
    ----------
    :class:`DeprecationWarning`
        If this attribute is deprecated.

    Raises
    ----------
    :exc:`AttributeError`
        If this attribute is unrecognized and thus erroneous.
    '''

    # Isolate imports to avoid polluting the module namespace.
    from beartype._util.module.utilmoddeprecate import deprecate_module_attr

    # Return the value of this deprecated attribute and emit a warning.
    return deprecate_module_attr(
        attr_deprecated_name=attr_deprecated_name,
        attr_deprecated_name_to_nondeprecated_name={
            'HintPep585Type': 'HintGenericSubscriptedType',
        },
        attr_nondeprecated_name_to_value=globals(),
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **object attribute utilities** (i.e., low-level callables handling
arbitrary attributes of objects in a general-purpose manner).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Callable,
    List,
    Optional,
)
# from beartype._cave._cavefast import MethodBoundInstanceDunderCType
from beartype._data.func.datafunc import OBJECT_SLOT_WRAPPERS
from beartype._data.typing.datatyping import DictStrToAny
from inspect import getattr_static

# ....................{ GETTERS                            }....................
def get_object_attrs_name_to_value_explicit(
    # Mandatory parameters.
    obj: object,

    # Optional parameters.
    obj_dir: Optional[List[str]] = None,
    predicate: Optional[Callable[[str, object], bool]] = None
) -> DictStrToAny:
    '''
    Dictionary mapping from the name to **explicit value** (i.e., value
    retrieved *without* implicitly calling the :func:`property`-decorated method
    implementing this attribute if this attribute is a property) of each
    attribute bound to the passed object whose name and/or value matches the
    passed predicate (in ascending lexicographic order of attribute name).

    This getter thus returns a dictionary such that each value is:

    * If the corresponding attribute is a **property** (i.e., method decorated
      by the standard :func:`property` decorator), the low-level data descriptor
      underlying this property rather than the high-level value returned by
      implicitly querying that data descriptor. Doing so avoids unexpected
      exceptions and is thus *significantly* safer.
    * Else, the value of this attribute as is.

    This getter is substantially safer than all known alternatives (e.g., the
    standard :func:`inspect.getmembers` getter), all of which implicitly call
    the low-level method implementing each high-level property of the passed
    object and hence raise exceptions when any such method raises exceptions. By
    compare, this getter *never* raises unexpected exceptions. Unless properties
    are of interest, callers are strongly encouraged to call this getter rather
    than unsafe alternatives.

    Caveats
    -------
    **This getter exhibits linear time complexity** :math:`O(n)` for :math:`n`
    the number of attributes transitively defined by the passed object
    (including both the type of that object and all superclasses of that type).
    This getter should thus be called with some measure of caution.

    **This getter only introspects attributes statically registered by the
    internal dictionary of this object** (e.g., ``__dict__`` in unslotted
    objects, ``__slots__`` in slotted objects). This getter thus silently
    ignores *all* attributes dynamically defined by the ``__getattr__()`` method
    or related runtime magic of this object.

    Parameters
    ----------
    obj : object
        Object to be introspected.
    obj_dir : Optional[List[str]]
        Either:

        * List of the names of all relevant attributes bound to this object.
          Callers may explicitly pass this list to either:

          * Consider only the proper subset of object attributes satisfying some
            external predicate. Doing so avoids the need to pass a ``predicate``
            callback, which can be surprisingly expensive in time to repeatedly
            call for each attribute.
          * Optimize away repeated calls to the :func:`dir` builtin, which are
            surprisingly expensive in both time and space.

        * :data:`None`, in which case this getter defaults this list to the
          names of *all* attributes bound to this object by calling the
          :func:`dir` builtin on this object.

        Defaults to :data:`None`.
    predicate: Optional[Callable[[str, object], bool]]
        Either:

        * Callable iteratively passed both the name and explicit value of each
          attribute bound to this object, returning :data:True` only if that
          name and/or value matches this predicate. This getter calls this
          callable for each attribute bound to this object and, if this callable
          returns :data:`True`, adds this name and explicit value to the
          returned dictionary as a new key-value pair. This predicate is
          expected to have a signature resembling:

          .. code-block:: python

             def predicate(attr_name: str, attr_value: object) -> bool: ...

        * :data:`None`, in which case this getter unconditionally adds *all*
          attributes bound to this object to this dictionary.

        Defaults to :data:`None`.

    Returns
    -------
    DictStrToAny
        Dictionary mapping from the name to explicit value of each attribute
        bound to the passed object whose name and/or value matches the passed
        predicate (in ascending lexicographic order of attribute name).
    '''
    assert obj_dir is None or isinstance(obj_dir, list), (
        f'{repr(obj_dir)} neither list of strings nor "None".')
    assert predicate is None or callable(predicate), (
        f'{repr(predicate)} neither callable nor "None".')

    # Dictionary mapping from the name of each attribute of the passed object
    # satisfying the passed predicate to the corresponding explicit value of
    # that attribute.
    attrs_name_to_value_explicit = None  # type: ignore[assignment]

    # If the caller passed *NO* list of attribute names, default this to the
    # list of *ALL* attribute names bound to this object.
    if obj_dir is None:
        obj_dir = dir(obj)
    # Else, the caller passed a list of attribute names.

    # If the caller passed a predicate...
    if predicate:
        # Initialize this dictionary to the empty dictionary.
        attrs_name_to_value_explicit = {}

        # Ideally, this function would be reimplemented in terms of the
        # iter_attrs_implicit_matching() function calling the canonical
        # inspect.getmembers() function. Dynamic inspection is surprisingly
        # non-trivial in the general case, particularly when virtual base
        # classes rear their diamond-studded faces. Moreover, doing so would
        # support edge-case attributes when passed class objects, including:
        # * Metaclass attributes of the passed class.
        #
        # Sadly, inspect.getmembers() internally accesses attributes via the
        # dangerous getattr() builtin rather than the safe
        # inspect.getattr_static() function. This function explicitly requires
        # the latter and hence *MUST* reimplement rather than defer to
        # inspect.getmembers(). (Sadness reigns.)
        #
        # For the same reason, the unsafe vars() builtin cannot be called
        # either. Since that builtin fails for builtin containers (e.g., "dict",
        # "list"), this is not altogether a bad thing.
        for attr_name in obj_dir:
            # Value of this attribute guaranteed to be statically rather than
            # dynamically retrieved. The getattr() builtin performs the latter,
            # dynamically calling this attribute's getter if this attribute is
            # a property. Since that call could conceivably raise unwanted
            # exceptions *AND* since this function explicitly ignores
            # properties, static attribute retrievable is unavoidable.
            attr_value = getattr_static(obj, attr_name)

            # If this attribute matches this predicate...
            if predicate(attr_name, attr_value):
                # Add the name and explicit value of this attribute to this
                # dictionary as a new key-vaue pair. Note that, due to the above
                # assignment, this iteration *CANNOT* reasonably be optimized
                # into a dictionary comprehension.
                attrs_name_to_value_explicit[attr_name] = attr_value
            # Else, this attribute fails to match this predicate. In this case,
            # silently ignore this attribute.
    # Else, the caller passed *NO* predicate. In this case...
    else:
        # Trivially define this dictionary via a dictionary comprehension.
        attrs_name_to_value_explicit = {
            attr_name: getattr_static(obj, attr_name)
            for attr_name in obj_dir
        }

    # Return this dictionary.
    return attrs_name_to_value_explicit


def get_object_methods_name_to_value_explicit(
    # Mandatory parameters.
    obj: object,

    # Optional parameters.
    obj_dir: Optional[List[str]] = None,
) -> DictStrToAny:
    '''
    Dictionary mapping from the name to **explicit value** (i.e., value
    retrieved *without* implicitly calling the :func:`property`-decorated method
    implementing this attribute if this attribute is a property) of each method
    bound to the passed object.

    Parameters
    ----------
    obj : object
        Object to be introspected.
    obj_dir : Optional[List[str]]
        See also the :func:`.get_object_attrs_name_to_value_explicit` getter.

    Caveats
    -------
    **This getter intentionally returns unbound pure-Python method functions
    rather than bound C-based method descriptors.** In theory, the latter
    approach would be marginally more useful. In practice, the standard
    :func:`.getattr_static` getter underlying this getter only supports the
    former approach. It is what it is.

    **This getter intentionally omits uncallable methods.** This includes most
    C-based method descriptors, most of which are uncallable depending on the
    version of the active Python interpreter. This *particularly* includes all
    C-based slot wrappers implicitly inherited by all classes from the root
    :class:`object` superclass (e.g., the :meth:`object.__str__` dunder method).
    The default implementations of slot wrappers have no intrinsic value in any
    meaningful context and only serve to obfuscate *actual* methods of
    general-purpose interest to most callers.

    Returns
    -------
    DictStrToAny
        Dictionary mapping from the name to explicit value of each methods bound
        to the passed object.

    Methods
    -------
    :func:`.get_object_attrs_name_to_value_explicit`
        Further details.
    '''

    # This is why we predicate, folks.
    return get_object_attrs_name_to_value_explicit(
        obj=obj,
        obj_dir=obj_dir,
        predicate=_is_object_attr_callable_not_object_slot_wrapper,
    )

# ....................{ PRIVATE ~ testers                  }....................
def _is_object_attr_callable_not_object_slot_wrapper(
    attr_name: str, attr_value: object) -> bool:
    '''
    Predicate suitable for passing as the ``predicate`` parameter to the
    :func:`.get_object_attrs_name_to_value_explicit` getter, returning
    :data:`True` only if the passed attribute value is both callable and *not*
    an **object slot wrappers** (i.e., low-level C-based callables bound to the
    root :class:`object` superclass providing mostly useless default
    implementations of popular dunder methods).
    '''
    # print(f'OBJECT_SLOT_WRAPPERS: {OBJECT_SLOT_WRAPPERS}')

    # If this attribute value is uncallable, return false immediately.
    if not callable(attr_value):
        return False
    # Else, this attribute value is callable.

    # Return true only if this callable is *NOT* an "object" slot wrapper.
    #
    # Note that:
    # * Although all standard callables are hashable, some user-defined
    #   callables are unhashable. Examples of unhashable callables include:
    #   * Unhashable pseudo-callables (i.e., unhashable objects whose classes
    #     define the __call__() dunder methods).
    # * The beartype._util.utilobject.is_object_hashable() tester is *NOT*
    #   necessarily safely importable here, due to chicken-and-egg issues. Ergo,
    #   we manually guard against unhashable callables.
    try:
        return attr_value not in OBJECT_SLOT_WRAPPERS
    # If doing so raises *ANY* exception, this callable is unhashable. However,
    # *ALL* "object" slot wrappers are hashable. It follows that this callable
    # is *NOT* an "object" slot wrapper. Despite being unhashable, this callable
    # *COULD* be of interest to the caller.
    except Exception:
        pass

    # Return true as a fallback.
    return True



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype** :pep:`484`-compliant **type variable utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep484.pep484typevar` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ getter                     }....................
def test_get_hint_pep484_typevar_bounded_constraints_or_none() -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484.pep484typevar.get_hint_pep484_typevar_bounded_constraints_or_none`
    tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep484Exception
    from beartype._data.typing.datatyping import T
    from beartype._util.hint.pep.proposal.pep484.pep484typevar import (
        get_hint_pep484_typevar_bounded_constraints_or_none)
    from beartype_test.a00_unit.data.pep.data_pep484 import (
        T_int,
        T_str_or_bytes,
    )
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this getter returns "None" for unbounded type variables.
    assert get_hint_pep484_typevar_bounded_constraints_or_none(T) is None

    # Assert this getter reduces bounded type variables to their upper bound.
    assert get_hint_pep484_typevar_bounded_constraints_or_none(T_int) is int

    # Union of all constraints parametrizing a constrained type variable,
    # reduced from that type variable.
    typevar_constraints_union = get_hint_pep484_typevar_bounded_constraints_or_none(
        T_str_or_bytes)

    # Assert this union contains all constraints parametrizing this variable.
    assert str   in typevar_constraints_union.__args__
    assert bytes in typevar_constraints_union.__args__

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed a non-type
    # variable.
    with raises(BeartypeDecorHintPep484Exception):
        get_hint_pep484_typevar_bounded_constraints_or_none(str)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype Decidedly Object-Oriented Runtime-checking (DOOR) class type hints**
(i.e., PEP-compliant type hints matching the hierarchy driving our
object-oriented type hint class hierarchy).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorsuper import TypeHint
from beartype.typing import (
    Tuple,
)

# ....................{ HINTS                              }....................
TupleTypeHints = Tuple[TypeHint, ...]
'''
PEP-compliant type hint matching a tuple of zero or more **type hint wrappers**
(i.e., :data:`.TypeHint` objects).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **object attribute caching utilities** (i.e., low-level callables
caching :mod:`beartype`-specific attributes describing user-defined
pure-Python functions, classes, and modules, typically by monkey-patching those
attributes directly into those objects).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilCacheObjectAttributeException
from beartype.typing import Dict
from beartype._cave._cavefast import (
    FunctionType,
    ModuleType,
)
from beartype._data.typing.datatyping import TypeException
from beartype._data.kind.datakindiota import SENTINEL
from functools import wraps
from threading import RLock

# ....................{ HINTS                              }....................
# Attribute-cachables are either...
ObjectAttrTypes = (
    # Pure-python functions *AR*...
    FunctionType,
    # Pure-python types *OR*...
    type,
    # Pure-python modules.
    ModuleType,
)
'''
Tuple of the types of all **attribute-cachables** (i.e., objects for which
arbitrary attributes may be safely cached, equivalent to objects accepted by the
:func:`.get_object_attr_cached_or_sentinel` getter and
:func:`.set_object_attr_cached` setter).
'''

# ....................{ GLOBALS                            }....................
OBJECT_ATTR_CACHE_LOCK = RLock()
'''
**Reentrant object attribute cache thread lock** (i.e., low-level thread locking
mechanism implemented as a highly efficient C extension, defined as a global for
reentrant reuse elsewhere as a context manager).

This lock is intentionally public, enabling external callers to synchronize
threading behaviour against the same object attribute caches synchronized by
this lock.

Note that a reentrant :class:`threading.RLock` is required, as higher-level
locked functions (e.g., :func:`.get_object_attr_cached_or_sentinel`) frequently
invoke lower-level locked functions (e.g.,
:func:`.get_type_attr_cached_or_sentinel`).
'''

# ....................{ CLEARERS                           }....................
#FIXME: Unit test us up, please.
def clear_object_attr_caches() -> None:
    '''
    Clear (i.e., empty) *all* private caches internally defined by this
    submodule, enabling callers to reset this submodule to its initial state.
    '''

    # Thread-safely...
    with OBJECT_ATTR_CACHE_LOCK:
        # Clear all private caches defined below.
        _MODULE_NAME_TO_ATTR_NAME_TO_VALUE.clear()

# ....................{ GETTERS                            }....................
#FIXME: Implement us up, please.
#FIXME: Docstring up the "attr_name_*" family of attributes, please.
def get_object_attr_cached_or_sentinel(
    # Mandatory parameters.
    obj: object,
    attr_name_if_obj_function: str,
    attr_name_if_obj_type_or_module: str,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilCacheObjectAttributeException,
    exception_prefix: str = '',
) -> object:
    '''
    **Memoized object attribute** (i.e., :mod:`beartype`-specific attribute
    memoizing the prior result of an expensive decision problem unique to the
    passed object) with the passed name previously cached about this object by a
    prior call to the :func:`.set_object_attr_cached` setter passed the same
    name if such a call occurred *or* the sentinel placeholder otherwise (e.g.,
    if no such call occurred).

    This getter is thread-safe with respect to the corresponding
    :func:`.set_object_attr_cached` setter.

    Caveats
    -------
    **This getter does not support arbitrary objects.** For safety, this getter
    *only* supports pure-Python functions, types, and modules. If the passed
    object is *any* type kind of object, this getter raises an exception of the
    passed type.

    Parameters
    ----------
    obj : object
        Object to be inspected.
    attr_name_if_obj_function : str
        Name of the memoized object attribute to be accessed on this object if
        this object is a pure-Python function. To avoid namespace collisions
        with both official dunder attributes (e.g., ``__module__``,
        ``__name__``) *and* third-party attributes also monkey-patched into this
        function, this name should be uniquified with a unique prefix and/or
        suffix (e.g., ``__beartype_attr__``, ``__attr_beartype__``).
    attr_name_if_obj_type_or_module : str
        Name of the memoized object attribute to be accessed on this object if
        this object is either a pure-Python type or module. Since this name is
        only internally used as the key of a private dictionary rather than an
        actual Python attribute, this name need *not* be uniquified with a
        unique prefix and/or suffix.
    exception_cls : TypeException, default: _BeartypeUtilCacheObjectAttributeException
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilCacheObjectAttributeException`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    object
        Either:

        * If a prior call to the :func:`.set_type_attr_cached` setter passed the
          same name previously monkey-patched this memoized type attribute into
          this type, the value of this attribute.
        * Else, the **sentinel placeholder** (i.e., :data:`.SENTINEL`).

    Raises
    ------
    exception_cls
        If this object is neither a pure-Python function, type, nor module.
    '''

    # Thread-safely...
    with OBJECT_ATTR_CACHE_LOCK:
        # Value of the attribute to be returned if previously cached on this
        # object *OR* the sentinel placeholder otherwise.
        attr_value: object = SENTINEL

        # If this object is a pure-Python function...
        #
        # Note that most objects of interest are pure-Python functions. This
        # common case is intentionally detected first as a microoptimization.
        if isinstance(obj, FunctionType):
            assert isinstance(attr_name_if_obj_function, str), (
                f'{repr(attr_name_if_obj_function)} not string.')

            # Value of this attribute if previously monkey-patched into this
            # function *OR* the sentinel placeholder otherwise.
            #
            # Note that attributes are intentionally monkey-patched into
            # functions rather than cached as nested dictionary entries as with
            # modules. Why? Because the latter approach would require each
            # function to have a unique name. You are now thinking: "B-b-but...
            # functions all have unique names! Don't they?" Sadly, the answer
            # is: "Nope." Property getters, setters, and deleters are *ALL*
            # pure-Python functions that share the same names. They're also
            # incredibly common. Because of the mere existence of @property
            # objects, attributes *MUST* instead be monkey-patched directly into
            # functions. Python do be like that.
            attr_value = getattr(obj, attr_name_if_obj_function, SENTINEL)
        # Else, this object is *NOT* a pure-Python function.
        #
        # If this object is a pure-Python type, defer to the lower-level getter
        # specific to types.
        #
        # Note that many objects of interest are pure-Python types. This common
        # case is intentionally detected next as a microoptimization.
        elif isinstance(obj, type):
            attr_value = get_type_attr_cached_or_sentinel(
                obj, attr_name_if_obj_type_or_module)
        # Else, this object is *NOT* a pure-Python type.
        #
        # If this object is a pure-Python module...
        #
        # Note that very few objects of interest are pure-Python modules. This
        # common case is intentionally detected last as a microoptimization.
        elif isinstance(obj, ModuleType):
            assert isinstance(attr_name_if_obj_type_or_module, str), (
                f'{repr(attr_name_if_obj_type_or_module)} not string.')

            # Avoid circular import dependencies.
            from beartype._util.module.utilmodget import get_module_name

            # Fully-qualified name of this module.
            module_name = get_module_name(obj)

            # Nested dictionary mapping from name to value of each previously
            # memoized attribute of this module if any *OR* "None" otherwise.
            attr_name_to_value = (
                _MODULE_NAME_TO_ATTR_NAME_TO_VALUE.get(module_name))

            # If no such nested dictionary exists, fallback to a new empty
            # nested dictionary.
            if not attr_name_to_value:
                attr_name_to_value = (
                    _MODULE_NAME_TO_ATTR_NAME_TO_VALUE[module_name]) = {}
            # Else, this nested dictionary has already been memoized.
            #
            # In either case, this nested dictionary now exists.

            # Value of this attribute if previously cached into this nested
            # dictionary *OR* the sentinel placeholder otherwise.
            attr_value = attr_name_to_value.get(
                attr_name_if_obj_type_or_module, SENTINEL)
        # Since this object is of an unknown type, arbitrary attributes *CANNOT*
        # be safely monkey-patched into this object; likewise, this object has
        # no unique identifier with which to cache arbitrary attributes inside
        # external datastores. This object is *NOT* cacheable. In this case...
        else:
            assert isinstance(exception_cls, type), (
                f'{repr(exception_cls)} not type.')
            assert isinstance(exception_prefix, str), (
                f'{repr(exception_prefix)} not string.')

            # Raise an exception.
            raise exception_cls(
                f'{exception_prefix}object {repr(obj)} neither '
                f'pure-Python function, class, nor module.'
            )

        # Return the value of this attribute.
        return attr_value


#FIXME: Unit test us up, please.
def get_type_attr_cached_or_sentinel(
    # Mandatory parameters.
    cls: type,
    attr_name: str,

    # Optional parameters.
    is_dirty: bool = False,
) -> object:
    '''
    **Memoized type attribute** (i.e., :mod:`beartype`-specific attribute
    memoizing the prior result of an expensive decision problem unique to the
    passed type) with the passed name previously monkey-patched into this type
    by a prior call to the :func:`.set_type_attr_cached` setter passed the same
    name if such a call occurred *or* the sentinel placeholder otherwise (e.g.,
    if no such call occurred).

    This getter is thread-safe with respect to the corresponding
    :func:`.set_type_attr_cached` setter.

    Caveats
    -------
    **Memoized type attributes are only accessible by calling this getter.**
    Memoized type attributes are *not* monkey-patched directly into types.
    Memoized type attributes are only monkey-patched indirectly into types.
    Specifically, the :func:`.set_type_attr_cached` setter monkey-patches
    memoized type attributes into pure-Python ``__sizeof__()`` dunder methods
    monkey-patched into types. Why? Safety. Monkey-patching attributes directly
    into types would conflict with user expectations, which expect class
    dictionaries to remain untrammelled by third-party decorators.

    Parameters
    ----------
    cls : type
        Type to be inspected.
    attr_name : str
        Unqualified basename of the memoized type attribute to be retrieved.
    is_dirty : bool, default: False
        :data:`True` only if the current cache entry for this attribute is
        **dirty** (i.e., stale, desynchronized), in which case this getter
        additionally **invalidates** (i.e., clears, removes) this dirty cache
        entry as a beneficial side-effect. Defaults to :data:`False`.

    Returns
    -------
    object
        Either:

        * If a prior call to the :func:`.set_type_attr_cached` setter passed the
          same name previously monkey-patched this memoized type attribute into
          this type, the value of this attribute.
        * Else, the **sentinel placeholder** (i.e., :data:`.SENTINEL`).
    '''
    assert isinstance(cls, type), f'{repr(cls)} not type.'
    assert isinstance(attr_name, str), f'{repr(attr_name)} not string.'
    assert isinstance(is_dirty, bool), f'{repr(is_dirty)} not boolean.'

    # Thread-safely...
    with OBJECT_ATTR_CACHE_LOCK:
        # __sizeof__() dunder method currently declared by this class, which the
        # set_type_attr_cached() setter has possibly wrapped with a pure-Python
        # __sizeof__() dunder method. Why? Tangential reasons that are obscure,
        # profane, and have *NOTHING* to do with the __sizeof__() dunder method
        # itself. Succinctly, we need a reasonably safe place to persist
        # @beartype-specific attributes pertaining to this class.
        #
        # Clearly, the obvious place would be this class itself. However, doing
        # so would fundamentally modify this class and thus *ALL* instances of
        # this class in an unexpected and thus possibly unsafe manner. Consider
        # common use cases like slots, introspection, pickling, and sizing.
        # Clearly, monkey-patching attributes into class dictionaries without
        # the explicit consent of class designers (i.e., users) is an
        # ill-advised approach.
        #
        # A less obvious but safer place is required. A method of this class
        # would be the ideal candidate; whereas everybody cares about object
        # attributes and thus class dictionaries, nobody cares about method
        # attributes. This is why @beartype safely monkey-patches attributes
        # into @beartype-decorated methods. However, which method? Most methods
        # are *NOT* guaranteed to exist across all possible classes. Adding a
        # new method to this class would be no better than adding a new
        # attribute to this class; both modify class dictionaries. Fortunately,
        # Python currently guarantees *ALL* classes to define at least 24 dunder
        # methods as of Python 3.11. How? Via the root "object" superclass.
        # Unfortunately, *ALL* of these methods are C-based and thus do *NOT*
        # directly support monkey-patching: e.g.,
        #     >>> class AhMahGoddess(object): pass
        #     >>> AhMahGoddess.__init__.__beartyped_cls = AhMahGoddess
        #     AttributeError: 'wrapper_descriptor' object has no attribute
        #     '__beartyped_cls'
        #
        # Fortunately, *ALL* of these methods may be wrapped by pure-Python
        # equivalents whose implementations defer to their original C-based
        # methods. Unfortunately, doing so slightly reduces the efficiency of
        # calling these methods. Fortunately, a subset of these methods are
        # rarely called under production workloads; slightly reducing the
        # efficiency of calling these methods is irrelevant to almost all use
        # cases. Of these, the most obscure, largely useless, poorly documented,
        # and single-use is the __sizeof__() dunder method -- which is only ever
        # called by the sys.getsizeof() utility function, which itself is only
        # ever called manually in a REPL or by third-party object sizing
        # packages. In short, __sizeof__() is perfect.
        cls_sizeof = cls.__sizeof__

        # If this method is *NOT* pure-Python, this method is C-based and thus
        # *CANNOT* possibly have been monkey-patched by a prior call to the
        # set_type_attr_cached() setter, which would have necessarily wrapped
        # this non-monkey-patchable C-based method with a monkey-patchable
        # pure-Python equivalent. In this case, return the sentinel placeholder.
        if not isinstance(cls_sizeof, FunctionType):
            return SENTINEL
        # Else, this method is pure-Python and thus *COULD* possibly have been
        # monkey-patched by a prior call to the set_type_attr_cached() setter.

        # Memoized type attribute cache (i.e., dictionary mapping from each type
        # in a type hierarchy passed to this setter to a nested dictionary
        # mapping from the name to value of each memoized type attribute cached
        # by a call to this setter) if the set_type_attr_cached() setter has
        # already been passed this type at least once *OR* "None" (i.e., if that
        # setter has yet to be passed this type). See that setter for details.
        type_to_attr_name_to_value = getattr(
            cls_sizeof, _TYPE_ATTR_CACHE_NAME, None)

        # If this cache does *NOT* exist, the passed type attribute *CANNOT*
        # possibly have been cached by a prior call to that setter. In this
        # case, return the sentinel placeholder.
        if not type_to_attr_name_to_value:
            return SENTINEL
        # Else, this cache exists. This type attribute *COULD* possibly have
        # been cached by a prior call to that setter.

        # Nested dictionary mapping from the name to value of each memoized type
        # attribute cached for this type by a prior call to that setter if this
        # nested dictionary exists *OR* "None" otherwise.
        attr_name_to_value = type_to_attr_name_to_value.get(cls)

        # If this nested dictionary has yet to be created, the passed type
        # attribute *CANNOT* possibly have been cached by a prior call to that
        # setter. In this case, return the sentinel placeholder.
        if not attr_name_to_value:
            return SENTINEL
        # Else, this nested dictionary. This type attribute *COULD* possibly
        # have been cached by a prior call to that setter.

        # Value of this type attribute cached by a prior call to that setter if
        # any *OR* the sentinel placeholder otherwise.
        attr_value = attr_name_to_value.get(attr_name, SENTINEL)

        # If...
        if (
            # The caller requests this attribute be marked "dirty" and thus
            # removed as an entry of this nested dictionary *AND*...
            is_dirty and
            # This attribute is an entry of this nested dictionary...
            attr_value is not SENTINEL
        ):
            # Remove this entry from this nested dictionary.
            del attr_name_to_value[attr_name]
        # Else, either the caller did not request this attribute to be
        # marked "dirty" *OR* this attribute has not yet been monkey-patched
        # into this function, preserve this attribute as is.

        # Return this value.
        return attr_value

# ....................{ SETTERS                            }....................
#FIXME: Unit test us up, please.
def set_object_attr_cached(
    # Mandatory parameters.
    obj: object,
    attr_name_if_obj_function: str,
    attr_name_if_obj_type_or_module: str,
    attr_value: object,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilCacheObjectAttributeException,
    exception_prefix: str = '',
) -> None:
    '''
    Monkey-patch the **memoized type attribute** (i.e., :mod:`beartype`-specific
    attribute memoizing the prior result of an expensive decision problem unique
    to the passed type) with the passed name and value this type.

    This setter is thread-safe with respect to the corresponding
    :func:`.get_object_attr_cached_or_sentinel` getter.

    Caveats
    -------
    **This setter does not support arbitrary objects.** For safety, this setter
    *only* supports pure-Python functions, types, and modules. If the passed
    object is *any* type kind of object, this setter raises an exception of the
    passed type.

    Parameters
    ----------
    cls : type
        Type to be inspected.
    attr_name_if_obj_function : str
        Name of the memoized object attribute to be accessed on this object if
        this object is a pure-Python function. To avoid namespace collisions
        with both official dunder attributes (e.g., ``__module__``,
        ``__name__``) *and* third-party attributes also monkey-patched into this
        function, this name should be uniquified with a unique prefix and/or
        suffix (e.g., ``__beartype_attr__``, ``__attr_beartype__``).
    attr_name_if_obj_type_or_module : str
        Name of the memoized object attribute to be accessed on this object if
        this object is either a pure-Python type or module. Since this name is
        only internally used as the key of a private dictionary rather than an
        actual Python attribute, this name need *not* be uniquified with a
        unique prefix and/or suffix.
    attr_value : object
        New value of this attribute.
    exception_cls : TypeException, default: _BeartypeUtilCacheObjectAttributeException
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilCacheObjectAttributeException`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.
    '''

    # Thread-safely...
    with OBJECT_ATTR_CACHE_LOCK:
        # If this object is a pure-Python function...
        #
        # Note that most objects of interest are pure-Python functions. This
        # common case is intentionally detected first as a microoptimization.
        if isinstance(obj, FunctionType):
            assert isinstance(attr_name_if_obj_function, str), (
                f'{repr(attr_name_if_obj_function)} not string.')

            # Monkey-patch the new value of this attribute into this function.
            setattr(obj, attr_name_if_obj_function, attr_value)
        # Else, this object is *NOT* a pure-Python function.
        #
        # If this object is a pure-Python type, defer to the lower-level getter
        # specific to types.
        #
        # Note that many objects of interest are pure-Python types. This common
        # case is intentionally detected next as a microoptimization.
        elif isinstance(obj, type):
            set_type_attr_cached(
                cls=obj,
                attr_name=attr_name_if_obj_type_or_module,
                attr_value=attr_value,
            )
        # Else, this object is *NOT* a pure-Python type.
        #
        # If this object is a pure-Python module...
        #
        # Note that very few objects of interest are pure-Python modules. This
        # common case is intentionally detected last as a microoptimization.
        elif isinstance(obj, ModuleType):
            assert isinstance(attr_name_if_obj_type_or_module, str), (
                f'{repr(attr_name_if_obj_type_or_module)} not string.')

            # Avoid circular import dependencies.
            from beartype._util.module.utilmodget import get_module_name

            # Fully-qualified name of this module.
            module_name = get_module_name(obj)

            # Nested dictionary mapping from name to value of each previously
            # memoized attribute of this module if any *OR* "None" otherwise.
            attr_name_to_value = (
                _MODULE_NAME_TO_ATTR_NAME_TO_VALUE.get(module_name))

            # If no such nested dictionary exists, fallback to a new empty
            # nested dictionary.
            if not attr_name_to_value:
                attr_name_to_value = (
                    _MODULE_NAME_TO_ATTR_NAME_TO_VALUE[module_name]) = {}
            # Else, this nested dictionary has already been memoized.
            #
            # In either case, this nested dictionary now exists.

            # Cache the new value of this attribute into this nested dictionary.
            attr_name_to_value[attr_name_if_obj_type_or_module] = attr_value
        # Since this object is of an unknown type, arbitrary attributes *CANNOT*
        # be safely monkey-patched into this object; likewise, this object has
        # no unique identifier with which to cache arbitrary attributes inside
        # external datastores. This object is *NOT* cacheable. In this case...
        else:
            assert isinstance(exception_cls, type), (
                f'{repr(exception_cls)} not type.')
            assert isinstance(exception_prefix, str), (
                f'{repr(exception_prefix)} not string.')

            # Raise an exception.
            raise exception_cls(
                f'{exception_prefix}object {repr(obj)} neither '
                f'pure-Python function, class, nor module.'
            )


#FIXME: Unit test us up, please.
def set_type_attr_cached(
    cls: type, attr_name: str, attr_value: object) -> None:
    '''
    Monkey-patch the **memoized type attribute** (i.e., :mod:`beartype`-specific
    attribute memoizing the prior result of an expensive decision problem unique
    to the passed type) with the passed name and value this type.

    This setter is thread-safe with respect to the corresponding
    :func:`.get_type_attr_cached_or_sentinel` getter.

    Caveats
    -------
    **Memoized type attributes are only mutatable by calling this setter.**
    Memoized type attributes are *not* monkey-patched directly into types.
    Memoized type attributes are only monkey-patched indirectly into types.
    Specifically, this setter monkey-patches memoized type attributes into
    pure-Python ``__sizeof__()`` dunder methods monkey-patched into types. Why?
    Safety. Monkey-patching attributes directly into types would conflict with
    user expectations, which expect class dictionaries to remain untrammelled by
    third-party decorators like :mod:`beartype`.

    Parameters
    ----------
    cls : type
        Type to be inspected.
    attr_name : str
        Unqualified basename of the memoized type attribute to be mutated.
    attr_value : object
        New value of this attribute.
    '''
    assert isinstance(cls, type), f'{repr(cls)} not type.'
    assert isinstance(attr_name, str), f'{repr(attr_name)} not string.'

    # Thread-safely...
    with OBJECT_ATTR_CACHE_LOCK:
        # __sizeof__() dunder method currently declared by this class. See the
        # get_type_attr_cached_or_sentinel() getter for details.
        cls_sizeof_old = cls.__sizeof__

        # If this method is already pure-Python, this method is already
        # monkey-patchable. In this case, monkey-patch this method directly.
        if isinstance(cls_sizeof_old, FunctionType):
            cls_sizeof = cls_sizeof_old  # pyright: ignore
        # Else, this method is *NOT* pure-Python, implying this method is
        # C-based and *NOT* monkey-patchable. In this case...
        else:
            # Avoid circular import dependencies.
            from beartype._util.cls.utilclsset import set_type_attr

            # New pure-Python __sizeof__() dunder method wrapping the original
            # C-based __sizeof__() dunder method declared by this class.
            @wraps(cls_sizeof_old)
            def cls_sizeof(self) -> int:
                return cls_sizeof_old(self)  # type: ignore[call-arg]

            # Replace the original C-based __sizeof__() dunder method with this
            # wrapper. For safety, we intentionally call our high-level
            # set_type_attr() setter rather than attempting to directly set this
            # attribute. The latter approach succeeds for standard pure-Python
            # mutable classes but catastrophically fails for non-standard
            # C-based immutable classes (e.g., "enum.Enum" subclasses).
            set_type_attr(cls, '__sizeof__', cls_sizeof)
        # Else, this method is already pure-Python.
        #
        # In any case, this method is now pure-Python and thus monkey-patchable.

        # Memoized type attribute cache (i.e., dictionary mapping from each type
        # in a type hierarchy passed to this setter to a nested dictionary
        # mapping from the name to value of each memoized type attribute cached
        # by a call to this setter) if this setter has already been passed this
        # type at least once *OR* "None" (i.e., if this setter has yet to be
        # passed this type).
        #
        # Ideally, this dictionary would *NOT* be required. Instead, this setter
        # would simply monkey-patch memoized type attributes directly into this
        # pure-Python __sizeof__() dunder method. Indeed, that overly simplistic
        # approach *DOES* work for a subset of cases: namely, if this type has
        # *NO* subclasses that are also passed to this setter. But if this type
        # his a subclass that is also passed to this setter, that approach would
        # cache the incorrect values. Why? Because subclasses of this type
        # inherit this pure-Python __sizeof__() dunder method and thus *ALL*
        # attributes monkey-patched by this setter into that method: e.g.,
        #     >>> class Superclass(): pass
        #     >>> def patch_sizeof(cls):
        #     ...     sizeof_old = cls.__sizeof__
        #     ...     def sizeof_new(self):
        #     ...         return sizeof_old(self)
        #     ...     cls.__sizeof__ = sizeof_new
        #     >>> Superclass.__sizeof__
        #     <method '__sizeof__' of 'object' objects>
        #     >>> patch_sizeof(Superclass)
        #     >>> Superclass.__sizeof__
        #     <function patch_sizeof.<locals>.sizeof_new at 0x7f1981393110>
        #
        #     >>> class Subclass(Superclass): pass
        #     >>> Subclass.__sizeof__
        #     <function patch_sizeof.<locals>.sizeof_new at 0x7f1981393110>
        #
        # Cache entries *MUST* thus be uniquified across type hierarchies.
        type_to_attr_name_to_value = getattr(
            cls_sizeof, _TYPE_ATTR_CACHE_NAME, None)

        # If *NO* memoized type attribute cache has been monkey-patched into
        # this pure-Python __sizeof__() dunder method yet, do so.
        if type_to_attr_name_to_value is None:
            type_to_attr_name_to_value = cls_sizeof._TYPE_ATTR_CACHE_NAME = {}  # type: ignore[attr-defined]
        # Else, a memoized type attribute cache has already been monkey-patched
        # into this pure-Python __sizeof__() dunder method.
        #
        # In either case, this cache now exists.

        # Nested dictionary mapping from the name to value of each memoized type
        # attribute cached for this type by a prior call to this setter if this
        # nested dictionary exists *OR* "None" otherwise.
        attr_name_to_value = type_to_attr_name_to_value.get(cls)

        # If this nested dictionary has yet to be created, do so.
        if attr_name_to_value is None:
            attr_name_to_value = type_to_attr_name_to_value[cls] = {}
        # Else, this nested dictionary has already been created.
        #
        # In either case, this nested dictionary now exists.

        # Cache this memoized type attribute into this nested dictionary. Phew!
        attr_name_to_value[attr_name] = attr_value  # type: ignore[index, assignment]

# ....................{ PRIVATE ~ globals                  }....................
_MODULE_NAME_TO_ATTR_NAME_TO_VALUE: Dict[str, Dict[str, object]] = {}
'''
Dictionary mapping from the fully-qualified name of each module associated with
one or more memoized module attributes to a nested dictionary mapping from the
name to value of each such attribute.
'''


_TYPE_ATTR_CACHE_NAME = '__beartype_type_attr_cache'
'''
Arbitrary :mod:`beartype`-specific name of the **memoized type attribute cache**
(i.e., dictionary mapping from each type in a type hierarchy passed to the
:func:`set_type_attr_cached` setter to a nested dictionary mapping from the name
to value of each memoized type attribute cached by a prior call to that setter)
monkey-patched into types passed to that setter.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype import hook** :pep:`526` **subpackage initialization submodule**
(i.e., data module mimicking real-world usage of various :func:`beartype.claw`
import hooks on :pep:`526`-compliant annotated variable assignments).
'''

# ....................{ IMPORTS                            }....................
from beartype import BeartypeConf
from beartype.claw import beartype_package
from beartype.roar import BeartypeValeLambdaWarning
from beartype_test.a00_unit.data.claw.intraprocess.hookable_package.pep.pep526 import (
    data_claw_pep526_raise)

# Subject this single module to a beartype import hook configured to silently
# ignore (rather than type-check) all PEP 526-compliant annotated assignments.
beartype_package(
    'beartype_test.a00_unit.data.claw.intraprocess.hookable_package.pep.pep526.data_claw_pep526_ignore',
    conf=BeartypeConf(claw_is_pep526=False),
)
from beartype_test.a00_unit.data.claw.intraprocess.hookable_package.pep.pep526 import (
    data_claw_pep526_ignore)

# Subject this single module to a beartype import hook configured to emit
# non-fatal warnings of an arbitrary beartype-specific warning type unlikely to
# arise via accidental change rather than raise fatal exceptions.
beartype_package(
    'beartype_test.a00_unit.data.claw.intraprocess.hookable_package.pep.pep526.data_claw_pep526_warn',
    conf=BeartypeConf(violation_door_type=BeartypeValeLambdaWarning),
)
from beartype_test.a00_unit.data.claw.intraprocess.hookable_package.pep.pep526 import (
    data_claw_pep526_warn)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator PEP-compliant type-checking code generator unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._check.code.codemain` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ pass : check               }....................
def test_make_check_code_decoration() -> None:
    '''
    Test the :func:`beartype._check.code.codemain.make_check_expr`
    function.
    '''

    # Defer test-specific imports.
    from beartype._check.code.codemain import make_check_expr
    from beartype._check.metadata.hint.hintsane import HintSane
    from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT

    # Assert this function generates identical code for identical hints and is
    # thus cached via memoization.
    #
    # Note that the make_check_expr() API expects the passed hint to already be
    # sanified and thus encapsulated by a "HintSign" object.
    assert (
        make_check_expr(HintSane(str), BEARTYPE_CONF_DEFAULT) is
        make_check_expr(HintSane(str), BEARTYPE_CONF_DEFAULT)
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **current package beartype import hook skippable package**
(i.e., data package blacklisted by one or more import hooks published by the
:mod:`beartype.claw` subpackage and thus expected to *not* be type-checked by
the :func:`beartype.beartype` decorator).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`591` unit tests.

This submodule unit tests :pep:`591` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_decor_pep591() -> None:
    '''
    Test :pep:`591` support implemented in the :func:`beartype.beartype`
    decorator.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from beartype.typing import Final
    from dataclasses import dataclass
    from pytest import raises

    # ..................{ LOCALS                             }..................
    @beartype
    @dataclass
    class WildSpirit(object):
        '''
        Arbitrary dataclass type-checked by :func:`beartype.beartype`.
        '''

        which_art_moving_everywhere: Final[str] = (
            'Destroyer and preserver; hear, oh hear!')
        '''
        Field annotated by the :obj:`typing.Final` type hint factory subscripted
        by an arbitrary child type hint, passed by the :func:`.dataclass`
        decorator to :meth:`.__init__`.
        '''

        heaven_and_ocean: Final = (
            b'Shook from the tangled boughs of Heaven and Ocean,')
        '''
        Field annotated by the unsubscripted :obj:`typing.Final` type hint
        factory, passed by the :func:`.dataclass` decorator to
        :meth:`.__init__`.
        '''

    # Arbitrary instance of this dataclass exercising all edge cases.
    destroyer_and_preserver = WildSpirit()

    # ..................{ PASS                               }..................
    # Assert this dataclass defines the expected attributes.
    assert destroyer_and_preserver.which_art_moving_everywhere == (
        'Destroyer and preserver; hear, oh hear!')
    assert destroyer_and_preserver.heaven_and_ocean == (
        b'Shook from the tangled boughs of Heaven and Ocean,')

    #FIXME: Refactor this to test that redefinition fails *AFTER* @beartype
    #fully supports PEP 591. *sigh*
    # Assert that attempting to redefine this final field currently silently
    # succeeds, despite technically violating PEP 591.
    destroyer_and_preserver.which_art_moving_everywhere = (
        "Loose clouds like earth's decaying leaves are shed,")

    # ..................{ FAIL                               }..................
    # Assert that attempting to instantiate an instance of this dataclass with a
    # parameter violating the corresponding type hint annotating the field of
    # the same name raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        WildSpirit(which_art_moving_everywhere=(
            b"Thou on whose stream, mid the steep sky's commotion,"))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-compliant **generic type hint utilities** (i.e.,
callables generically applicable to :pep:`484`-compliant generic classes).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep484Exception
from beartype.typing import Generic
from beartype._data.cls.datacls import TYPES_PEP484544_GENERIC
from beartype._data.typing.datatypingport import Hint
from beartype._data.typing.datatyping import TypeException
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.cls.utilclstest import is_type_subclass
from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11

# ....................{ TESTERS                            }....................
#FIXME: Unit test us up, please.
def is_hint_pep484_generic_subbed(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed object is a :pep:`484`-compliant
    **subscripted generic** (i.e., object subscripted by one or more child type
    hints originating from a :class:`typing.Generic` subclass, typically due to
    originally subclassing at least one PEP-compliant type hint defined by the
    standard :mod:`typing` module).

    This tester is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as the implementation trivially reduces to
    an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a :pep:`484`-compliant subscripted
        generic.

    See Also
    --------
    :func:`.is_hint_pep484_generic_unsubbed`
        Further details.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import get_hint_pep_origin_or_none

    # Arbitrary object originating this hint if any *OR* "None" otherwise.
    hint_origin = get_hint_pep_origin_or_none(hint)

    # Return true only if...
    return (
        # Some object originates this hint *AND*...
        hint_origin is not None and
        # This origin object is an unsubscripted generic type, which would then
        # imply this hint to be a subscripted generic. If this strikes you as
        # insane, you're not alone
        is_hint_pep484_generic_unsubbed(hint_origin)
    )


#FIXME: Unit test us up, please.
def is_hint_pep484_generic_unsubbed(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed object is a :pep:`484`-compliant
    **unsubscripted generic** (i.e., :class:`typing.Generic` subclass, typically
    due to originally subclassing at least one PEP-compliant type hint defined
    by the standard :mod:`typing` module).

    This tester returns :data:`True` only if this object was originally defined
    as a class subclassing a combination of:

    * At least one of:

      * The :pep:`484`-compliant :mod:`typing.Generic` superclass.
      * The :pep:`544`-compliant :mod:`typing.Protocol` superclass.

    * Zero or more non-class :mod:`typing` pseudo-superclasses (e.g.,
      ``typing.List[int]``).
    * Zero or more other standard superclasses.

    This tester is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as the implementation trivially reduces to
    an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a :pep:`484`-compliant unsubscripted
        generic.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import get_hint_pep_args

    # Return true only if this hint is a subclass of the "typing.Generic"
    # superclass, in which case this hint is a generic.
    #
    # Note that:
    # * This test is robust against edge cases, as the "typing" module
    #   guarantees all user-defined classes subclassing one or more "typing"
    #   pseudo-superclasses to subclass the "typing.Generic" abstract base class
    #   (ABC) regardless of whether those classes did so explicitly. How? By
    #   type erasure, of course, the malignant gift that keeps on giving:
    #       >>> import typing as t
    #       >>> class MuhList(t.List): pass
    #       >>> MuhList.__orig_bases__
    #       (typing.List)
    #       >>> MuhList.__mro__
    #       (__main__.MuhList, list, typing.Generic, object)
    # * This issubclass() call implicitly performs a surprisingly
    #   inefficient search over the method resolution order (MRO) of all
    #   superclasses of this hint. In theory, the cost of this search might
    #   be circumventable by observing that this ABC is expected to reside
    #   at the second-to-last index of the tuple exposing this MRO far all
    #   generics by virtue of fragile implementation details violating
    #   privacy encapsulation. In practice, this codebase is already
    #   fragile enough.
    # * The following logic superficially appears to implement the same
    #   test *WITHOUT* the onerous cost of a search:
    #       return len(get_hint_pep484_generic_bases_unerased_or_none(hint)) > 0
    #   Why didn't we opt for that, then? Because this tester is routinely
    #   passed objects that *CANNOT* be guaranteed to be PEP-compliant.
    #   Indeed, the high-level is_hint_pep() tester establishing the
    #   PEP-compliance of arbitrary objects internally calls this
    #   lower-level tester to do so. Since the
    #   get_hint_pep484_generic_bases_unerased_or_none() getter internally
    #   reduces to returning the tuple of the general-purpose
    #   "__orig_bases__" dunder attribute formalized by PEP 560, testing
    #   whether that tuple is non-empty or not in no way guarantees this
    #   object to be a PEP-compliant generic.
    return (
        # If the active Python interpreter targets Python >= 3.11, a subclass
        # test suffices to detect a PEP 484-compliant unsubscripted generic;
        is_type_subclass(hint, Generic)  # type: ignore[arg-type]
        if IS_PYTHON_AT_LEAST_3_11 else
        # Else, the active Python interpreter targets Python <= 3.10. In this
        # case, a simple subclass test does *NOT* suffice to detect a PEP
        # 484-compliant unsubscripted generic. Why? Because Python <= 3.10
        # implements PEP 484-compliant subscripted generics as types! But
        # PEP 484-compliant unsubscripted generics are also types, of course:
        #     $ python3.10
        #     >>> from typing import List
        #     >>> class MuhGeneric(List): pass
        #     >>> isinstance(MuhGeneric, type)
        #     True  # <-- good
        #     >>> isinstance(MuhGeneric[str], type)
        #     True  # <-- *BAD*
        #
        #     $ python3.11
        #     >>> from typing import List
        #     >>> class MuhGeneric(List): pass
        #     >>> isinstance(MuhGeneric, type)
        #     True  # <-- good
        #     >>> isinstance(MuhGeneric[str], type)
        #     False  # <-- good
        #
        # Disambiguating this edge case requires also detecting whether this PEP
        # 484-compliant generic is subscripted by one or more child hints.
        (
            # This hint is a subclass of the PEP 484-compliant "typing.Generic"
            # superclass *AND*...
            is_type_subclass(hint, Generic) and  # type: ignore[arg-type]
            # This PEP 484-compliant generic is unsubscripted.
            not get_hint_pep_args(hint)
        )
    )

# ....................{ GETTERS                            }....................
@callable_cached
def get_hint_pep484_generic_bases_unerased(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep484Exception,
    exception_prefix: str = '',
) -> tuple:
    '''
    Tuple of all unerased :mod:`typing` **pseudo-superclasses** (i.e.,
    :mod:`typing` objects originally listed as superclasses prior to their
    implicit type erasure under :pep:`560`) of the passed :pep:`484`-compliant
    **generic** (i.e., either subscripted or unsubscripted type subclassing at
    least one non-class :mod:`typing` object).

    This getter is memoized for efficiency.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.
    exception_cls : TypeException
        Type of exception to be raised. Defaults to
        :exc:`.BeartypeDecorHintPep484Exception`.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    tuple
        Tuple of the one or more unerased pseudo-superclasses of this
        :mod:`typing` generic. Specifically:

        * If this generic defines an ``__orig_bases__`` dunder instance
          variable, the value of that variable as is.
        * Else, the value of the ``__mro__`` dunder instance variable stripped
          of all ignorable classes conveying *no* semantic meaning, including:

          * This generic itself.
          * The :class:`object` root superclass.

    Raises
    ------
    exception_cls
        If this hint is either:

        * *Not* a :mod:`typing` generic.
        * A :mod:`typing` generic that erased *none* of its superclasses but
          whose method resolution order (MRO) lists strictly less than four
          classes. Valid :pep:`484`-compliant generics should list at least
          four classes, including (in order):

          #. This class itself.
          #. The one or more :mod:`typing` objects directly subclassed by this
             generic.
          #. The :class:`typing.Generic` superclass.
          #. The zero or more non-:mod:`typing` superclasses subsequently
             subclassed by this generic (e.g., :class:`abc.ABC`).
          #. The :class:`object` root superclass.

    See Also
    --------
    :func:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget.get_hint_pep484585_generic_bases_unerased`
        Further details.
    '''

    #FIXME: This tuple appears to be implemented erroneously -- at least under
    #Python 3.7, anyway. Although this tuple is implemented correctly for the
    #common case of user-defined types directly subclassing "typing" types,
    #this tuple probably is *NOT* implemented correctly for the edge case of
    #user-defined types indirectly subclassing "typing" types: e.g.,
    #    >>> import collections.abc, typing
    #    >>> T = typing.TypeVar('T')
    #    >>> class Direct(collections.abc.Sized, typing.Generic[T]): pass
    #    >>> Direct.__orig_bases__
    #    (collections.abc.Sized, typing.Generic[~T])
    #    >>> class Indirect(collections.abc.Container, Direct): pass
    #    >>> Indirect.__orig_bases__
    #    (collections.abc.Sized, typing.Generic[~T])
    #
    #*THAT'S COMPLETELY INSANE.* Clearly, their naive implementation failed to
    #account for actual real-world use cases.
    #
    #On the bright side, the current implementation prevents us from actually
    #having to perform a breadth-first traversal of all original superclasses
    #of this class in method resolution order (MRO). On the dark side, it's
    #pants-on-fire balls -- but there's not much we can do about that. *sigh*
    #
    #If we ever need to perform that breadth-first traversal, resurrect this:
    #
    #    # If this class was *NOT* subject to type erasure, reduce to a noop.
    #    if not hint_bases:
    #        return hint_bases
    #
    #    # Fixed list of all typing super attributes to be returned.
    #    superattrs = acquire_fixed_list(FIXED_LIST_SIZE_MEDIUM)
    #
    #    # 0-based index of the last item of this list.
    #    superattrs_index = 0
    #
    #    # Fixed list of all transitive superclasses originally listed by this
    #    # class iterated in method resolution order (MRO).
    #    hint_orig_mro = acquire_fixed_list(FIXED_LIST_SIZE_MEDIUM)
    #
    #    # 0-based indices of the current and last items of this list.
    #    hint_orig_mro_index_curr = 0
    #    hint_orig_mro_index_last = 0
    #
    #    # Initialize this list with the tuple of all direct superclasses of this
    #    # class, which iteration then expands to all transitive superclasses.
    #    hint_orig_mro[:len(hint_bases)] = hint_bases
    #
    #    # While the heat death of the universe has been temporarily forestalled...
    #    while (True):
    #        # Currently visited superclass of this class.
    #        hint_base = hint_orig_mro[hint_orig_mro_index_curr]
    #
    #        # If this superclass is a typing attribute...
    #        if is_hint_pep_type_typing(hint_base):
    #            # Avoid inserting this attribute into the "hint_orig_mro" list.
    #            # Most typing attributes are *NOT* actual classes and those that
    #            # are have no meaningful public superclass. Ergo, iteration
    #            # terminates with typing attributes.
    #            #
    #            # Insert this attribute at the current item of this list.
    #            superattrs[superattrs_index] = hint_base
    #
    #            # Increment this index to the next item of this list.
    #            superattrs_index += 1
    #
    #            # If this class subclasses more than the maximum number of "typing"
    #            # attributes supported by this function, raise an exception.
    #            if superattrs_index >= FIXED_LIST_SIZE_MEDIUM:
    #                raise BeartypeDecorHintPep560Exception(
    #                    '{} PEP type {!r} subclasses more than '
    #                    '{} "typing" types.'.format(
    #                        exception_prefix,
    #                        hint,
    #                        FIXED_LIST_SIZE_MEDIUM))
    #        # Else, this superclass is *NOT* a typing attribute. In this case...
    #        else:
    #            # Tuple of all direct superclasses originally listed by this class
    #            # prior to PEP 484 type erasure if any *OR* the empty tuple
    #            # otherwise.
    #            hint_base_bases = getattr(hint_base, '__orig_bases__')
    #
    #            #FIXME: Implement breadth-first traversal here.
    #
    #    # Tuple sliced from the prefix of this list assigned to above.
    #    superattrs_tuple = tuple(superattrs[:superattrs_index])
    #
    #    # Release and nullify this list *AFTER* defining this tuple.
    #    release_fixed_list(superattrs)
    #    del superattrs
    #
    #    # Return this tuple as is.
    #    return superattrs_tuple
    #
    #Also resurrect this docstring snippet:
    #
    #    Raises
    #    ----------
    #    BeartypeDecorHintPep560Exception
    #        If this object defines the ``__orig_bases__`` dunder attribute but that
    #        attribute transitively lists :data:`FIXED_LIST_SIZE_MEDIUM` or more :mod:`typing`
    #        attributes.
    #
    #Specifically:
    #  * Acquire a fixed list of sufficient size (e.g., 64). We probably want
    #    to make this a constant in "utilcachelistfixedpool" for reuse
    #    everywhere, as this is clearly becoming a common idiom.
    #  * Slice-assign "__orig_bases__" into this list.
    #  * Maintain two simple 0-based indices into this list:
    #    * "bases_index_curr", the current base being visited.
    #    * "bases_index_last", the end of this list also serving as the list
    #      position to insert newly discovered bases at.
    #  * Iterate over this list and keep slice-assigning from either
    #    "__orig_bases__" (if defined) or "__mro__" (otherwise) into
    #    "list[bases_index_last:len(__orig_bases__)]". Note that this has the
    #    unfortunate disadvantage of temporarily iterating over duplicates,
    #    but... *WHO CARES.* It still works and we subsequently
    #    eliminate duplicates at the end.
    #  * Return a frozenset of this list, thus implicitly eliminating
    #    duplicate superclasses.
    #FIXME: Actually, a simpler implementation could now be defined in terms of
    #the new iter_hint_pep560_bases_unerased() iterator. Still, let's
    #avoid doing so until we *ABSOLUTELY* must. *sigh*

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
        get_hint_pep484585_generic_type_or_none)

    # Either the unsubscripted generic reduced from this possibly subscripted
    # generic if this hint is a generic *OR* "None". Specifically, either:
    # * If this generic if unsubscripted, this unsubscripted generic as is.
    # * If this generic if subscripted, the unsubscripted generic underlying
    #   this subscripted generic.
    # * If this generic is *NOT* actually a generic, "None".
    hint = get_hint_pep484585_generic_type_or_none(hint)  # pyright: ignore

    # If this hint is *NOT* a PEP 484- or 544-compliant unsubscripted generic,
    # raise an exception.
    if not is_hint_pep484_generic_unsubbed(hint):
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} not '
            f'PEP 484 or 585 unsubscripted generic.'
        )
    # Else, this hint is a PEP 484- or 544-compliant unsubscripted generic.
    #
    # If this hint is a PEP 484- or 544-compliant generic superclass in either
    # unsubscripted form (e.g., "typing.Generic", "typing.Protocol") or
    # subscripted form (e.g., "typing.Generic[T]", "typing.Protocol[T]"), then
    # (by definition) this superclass subclasses *NO* pseudo-superclasses. In
    # this case, trivially return the empty tuple.

    #FIXME: Unit test this edge case up, please.
    elif hint in TYPES_PEP484544_GENERIC:
        return ()
    # Else, this hint is *NOT* a PEP 484- or 544-compliant generic superclass.

    # Unerased pseudo-superclasses of this generic if any *OR* "None" otherwise
    # (e.g., if this generic is a single-inherited protocol).
    hint_bases = getattr(hint, '__orig_bases__', None)

    # If this generic erased its superclasses, return these superclasses as is.
    if hint_bases is not None:
        return hint_bases
    # Else, this generic erased *NONE* of its superclasses. These superclasses
    # *MUST* by definition be unerased and thus safely returnable as is.

    # Unerased superclasses of this generic defined by the method resolution
    # order (MRO) for this generic.
    hint_bases = hint.__mro__  # pyright: ignore

    # If this MRO lists strictly less than four classes, raise an exception.
    # The MRO for any unerased generic should list at least four classes:
    # 1. This class itself.
    # 2. The one or more "typing" objects directly subclassed by this generic.
    # 3. The "typing.Generic" superclass. Note that this superclass is typically
    #    but *NOT* necessarily the second-to-last superclass. Since this ad-hoc
    #    heuristic is *NOT* an actual constraint, we intentionally avoid
    #    asserting this to be the case. An example in which "typing.Generic" is
    #    *NOT* the second-to-last superclass is:
    #        >>> from abc import ABC
    #        >>> from typing import Protocol
    #        >>> class ProtocolCustomSuperclass(Protocol): pass
    #        >>> class ProtocolCustomABC(ProtocolCustomSuperclass, ABC): pass
    #        >>> print(ProtocolCustomABC.__mro__)
    #        (<class '__main__.ProtocolCustomABC'>,
    #         <class '__main__.ProtocolCustomSuperclass'>,
    #         <class 'typing.Protocol'>,
    #         <class 'typing.Generic'>,
    #         <class 'abc.ABC'>,
    #         <class 'object'>)
    # 4. The "object" root superclass.
    if len(hint_bases) < 4:
        raise exception_cls(
            f'{exception_prefix}PEP 484 generic {repr(hint)} '
            f'subclasses less than four superclasses {repr(hint_bases)}.'
        )
    # Else, this MRO lists at least four classes.
    #
    # If any class listed by this MRO fails to comply with the above
    # expectations, raise an exception.
    elif hint_bases[0] is not hint:
        raise exception_cls(
            f'{exception_prefix}PEP 484 generic {repr(hint)} '
            f'first superclass {repr(hint_bases[0])} != {repr(hint)}.'
        )
    elif hint_bases[-1] is not object:
        raise exception_cls(
            f'{exception_prefix}PEP 484 generic {repr(hint)} '
            f'last superclass {repr(hint_bases[-1])} != {repr(object)}.'
        )
    # Else, all classes listed by this MRO comply with the above expectations.

    # Return a slice of this tuple preserving *ONLY* the non-ignorable
    # superclasses listed by this tuple for conformance with the tuple returned
    # by this getter from the "__orig_bases__", which similarly lists *ONLY*
    # non-ignorable superclasses. Specifically, strip from this tuple:
    # * This class itself.
    # * The "object" root superclass.
    #
    # Ideally, the ignorable "(beartype.|)typing.(Generic|Protocol)"
    # superclasses would also be stripped. Sadly, as exemplified by the above
    # counter-example, those superclasses are *NOT* guaranteed to occupy the
    # third- and second-to-last positions (respectively) of this tuple. Ergo,
    # stripping these superclasses safely would require an inefficient
    # iterative O(n) search across this tuple for those superclasses. Instead,
    # we defer ignoring these superclasses to the caller -- which necessarily
    # already (and hopefully efficiently) ignores ignorable superclasses.
    return hint_bases[1:-1]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **shell singletons** (i.e., magic constants pertaining to the
parent shell encapsulating the active Python interpreter, including the names of
:mod:`beartype`-specific environment variables officially recognized by
:mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Dict,
    Optional,
)

# ....................{ VARS ~ conf                        }....................
# @beartype-specific environment variables configuring beartype configurations
# (i.e., "beartype.BeartypeConf" instances).

SHELL_VAR_CONF_IS_COLOR_NAME = 'BEARTYPE_IS_COLOR'
'''
Name of the **color configuration environment variable** (i.e.,
:mod:`beartype`-specific environment variable officially recognized by
:mod:`beartype` as globally configuring the value of the
:attr:`beartype.BeartypeConf.is_color` tri-state boolean).
'''


SHELL_VAR_CONF_IS_COLOR_VALUE_TO_OBJ: Dict[str, Optional[bool]] = {
    'True': True,
    'False': False,
    'None': None,
}
'''
Dictionary mapping from each permissible string value for the **color
configuration environment variable** (i.e., whose name is
:data:`.CONF_IS_COLOR_NAME`) to the corresponding value of the
:attr:`beartype.BeartypeConf.is_color` tri-state boolean.
'''
